from datetime import datetime
import traceback
from io import BytesIO
import json
from ipaddress import IPv6Address
import random
from time import time
from threading import Lock

import struct
import socket
import selectors
import logging

from ..humans import human
from ..consensus import (
    validate_no_duplicate_output_references_in_transactions,
    validate_non_coinbase_transaction_by_itself,
    validate_non_coinbase_transaction_in_coinstate,
    ValidateTransactionError,
)
from .params import (
    TIME_BETWEEN_CONNECTION_ATTEMPTS,
    PORT,
    MAX_MESSAGE_SIZE,
    MAX_IBD_PEERS,
    IBD_PEER_TIMEOUT,
    GET_BLOCKS_INVENTORY_SIZE,
    GET_PEERS_INTERVAL,
    SWITCH_TO_ACTIVE_MODE_TIMEOUT,
    EMPTY_INVENTORY_BACKOFF,
)
from .messages import (
    SupportedVersion,
    MessageHeader,
    Message,
    Peer,
    PeersMessage,
    HelloMessage,
    GetBlocksMessage,
    GetDataMessage,
    GetPeersMessage,
    DataMessage,
    DATA_BLOCK,
    DATA_TRANSACTION,
    InventoryMessage,
    InventoryItem,
)
from .utils import get_recent_block_heights
from ..utils import block_filename
from ..__version__ import __version__


logger = logging.getLogger("skepticoin.networking")

LISTENING_SOCKET = "LISTENING_SOCKET"
IRRELEVANT = "IRRELEVANT"
MAGIC = b'MAJI'

INCOMING = "INCOMING"
OUTGOING = "OUTGOING"


def _new_context():
    # just a random number will do; this is used for debugging only.
    return random.randrange(1 << 64)


class Manager:
    def step(self, current_time):
        raise NotImplementedError


class NetworkManager(Manager):

    def __init__(self):
        self.my_addresses = set()
        self.connected_peers = {}
        self.disconnected_peers = {}

    def _sanity_check(self):
        for key in self.disconnected_peers:
            if key in self.connected_peers:
                raise Exception("this shouldn't happen %s" % (key,))

    def step(self, current_time):
        # logger.info("NetworkManager.step()")
        self._sanity_check()

        for disconnected_peer in list(self.disconnected_peers.values()):
            if (disconnected_peer.direction == OUTGOING and
                (disconnected_peer.host, disconnected_peer.port) not in self.my_addresses and
                    disconnected_peer.is_time_to_connect(current_time)):

                disconnected_peer.last_connection_attempt = current_time

                local_peer.start_outgoing_connection(disconnected_peer)

        for peer in list(self.connected_peers.values()):
            peer.step(current_time)

    def handle_peer_connected(self, remote_peer):
        logger.info("%15s NetworkManager.handle_peer_connected()" % remote_peer.host)

        key = (remote_peer.host, remote_peer.port, remote_peer.direction)
        if key in self.connected_peers:
            logger.warning("%15s duplicate peer %s" % (remote_peer.host, key))
            local_peer.disconnect(self.connected_peers[key], "duplicate")  # just drop the existing one

        self._sanity_check()
        self.connected_peers[key] = remote_peer
        if key in self.disconnected_peers:
            del self.disconnected_peers[key]
        self._sanity_check()

    def handle_peer_disconnected(self, remote_peer):
        logger.info("%15s NetworkManager.handle_peer_disconnected()" % remote_peer.host)

        key = (remote_peer.host, remote_peer.port, remote_peer.direction)

        self._sanity_check()
        if remote_peer.direction == OUTGOING:
            self.disconnected_peers[key] = remote_peer
        del self.connected_peers[key]
        self._sanity_check()

    def get_active_peers(self):
        return [p for p in self.connected_peers.values() if p.hello_sent and p.hello_received]

    def update_peer_db(self, remote_peer):
        if remote_peer.direction != OUTGOING:
            return

        # TSTTCPW... really quite barebones like this :-D
        db = [tuple(li) for li in json.loads(open("peers.json").read())]

        tup = (remote_peer.host, remote_peer.port, remote_peer.direction)
        if tup not in db:
            db.append(tup)

        with open("peers.json", "w") as f:
            json.dump(db, f, indent=4)

    def broadcast_block(self, block):
        logger.info("%15s ChainManager.broadcast_block(%s)" % ("", human(block.hash())))
        message = DataMessage(DATA_BLOCK, block)
        for peer in list(self.connected_peers.values()):
            try:
                # try/except b/c .send_message might try to set the selector for a just-closed sock to writing
                peer.send_message(message)
            except OSError:  # e.g. ConnectionRefusedError, "Bad file descriptor"
                pass
            except (ValueError, KeyError):  # seen in the wild for selector problems; should be more exactly matched tho
                pass

    def broadcast_transaction(self, transaction):
        message = DataMessage(DATA_TRANSACTION, transaction)
        for peer in list(self.connected_peers.values()):
            try:
                # try/except b/c .send_message might try to set the selector for a just-closed sock to writing
                peer.send_message(message)
            except OSError:  # e.g. ConnectionRefusedError, "Bad file descriptor"
                pass
            except (ValueError, KeyError):  # seen in the wild for selector problems; should be more exactly matched tho
                pass


def inventory_batch_handled(peer):
    """Has the full loop GetBlocks -> Inventory -> GetData (n times) -> Data (n times) been completed?"""
    return not peer.waiting_for_inventory and peer.inventory_messages == []


class ChainManager(Manager):

    def __init__(self, current_time):
        self.lock = Lock()
        self.coinstate = None
        self.actively_fetching_blocks_from_peers = []
        self.started_at = current_time
        self.transaction_pool = []

    def step(self, current_time):
        if not self.should_actively_fetch_blocks(current_time):
            return  # no manual action required, blocks expected to be sent to us instead.

        ibd_candidates = [
            peer for peer in local_peer.network_manager.get_active_peers()
            if current_time > peer.last_empty_inventory_response_at + EMPTY_INVENTORY_BACKOFF
        ]

        if len(ibd_candidates) == 0:
            return

        # TODO note that if a peer disconnects, it is never removed from actively_fetching_blocks_from_peers, which
        # means that your IBD will be stuck until it times out. Not the best but it will recover eventually at least.
        self.actively_fetching_blocks_from_peers = [
            (timeout_at, p)
            for (timeout_at, p) in self.actively_fetching_blocks_from_peers
            if current_time < timeout_at and
            not inventory_batch_handled(p)]

        if len(self.actively_fetching_blocks_from_peers) > MAX_IBD_PEERS:
            return

        get_blocks_message = self.get_get_blocks_message()

        # TODO once MAX_IBD_PEERS > 1... pick one that you haven't picked before? potentially :-D
        remote_peer = random.choice(ibd_candidates)
        remote_peer.waiting_for_inventory = True
        self.actively_fetching_blocks_from_peers.append((current_time + IBD_PEER_TIMEOUT, remote_peer))
        remote_peer.send_message(get_blocks_message)

        # timeout is only implemented half-baked: it currently only limits when an new GetBlocksMessage will be sent;
        # but doesn't take the timed-out peer out of the loop that it's already in. This is not necessarily a bad thing.

    def should_actively_fetch_blocks(self, current_time):
        return (
            (current_time > self.coinstate.head().timestamp + SWITCH_TO_ACTIVE_MODE_TIMEOUT)
            or (current_time <= self.started_at + 60)  # always sync w/ network right after restart
            or (current_time % 60 == 0)  # on average, every 1 minutes, do a network resync explicitly
        )

    def set_coinstate(self, coinstate):
        with self.lock:
            logger.info("%15s ChainManager.set_coinstate(%s)" % ("", coinstate))
            self.coinstate = coinstate
            self._cleanup_transaction_pool_for_coinstate(coinstate)

    def add_transaction_to_pool(self, transaction):
        with self.lock:
            logger.info("%15s ChainManager.add_transaction_to_pool(%s)" % ("", human(transaction.hash())))

            try:
                validate_non_coinbase_transaction_by_itself(transaction)
                validate_non_coinbase_transaction_in_coinstate(
                    transaction, self.coinstate.current_chain_hash, self.coinstate)

                # Horribly inefficiently implemented (AKA 'room for improvement)
                validate_no_duplicate_output_references_in_transactions(self.transaction_pool + [transaction])

                #  we don't do validate_no_duplicate_transactions here (assuming it's just been done before
                #  add_transaction_to_pool).

            except ValidateTransactionError as e:
                # TODO: dirty hack at this particular point... to allow for e.g. out-of-order transactions to not take
                # down the whole peer, but this should more specifically match for a short list of OK problems.
                logger.info("%15s INVALID transaction %s" % ("", str(e)))
                print("Invalid Transaction")
                with open("/tmp/%s.transaction" % human(transaction.hash()), 'wb') as f:
                    f.write(transaction.serialize())

                return False  # not successful

            self.transaction_pool.append(transaction)

        return True  # successfully added

    def get_state(self):
        with self.lock:
            return self.coinstate, self.transaction_pool

    def _cleanup_transaction_pool_for_coinstate(self, coinstate):
        # This is really the simplest (though not most efficient mechanism): simply remove now-invalid transactions from
        # the pool
        def is_valid(transaction):
            try:
                # validate_non_coinbase_transaction_by_itself(transaction) Not needed, this never changes
                validate_non_coinbase_transaction_in_coinstate(
                    transaction, self.coinstate.current_chain_hash, self.coinstate)

                # Not needed either, this never becomes True if it was once False
                # validate_no_duplicate_output_references_in_transactions(self.transaction_pool + [transaction])
                return True
            except ValidateTransactionError:
                return False

        self.transaction_pool = [t for t in self.transaction_pool if is_valid(t)]

    def get_get_blocks_message(self):
        heights = get_recent_block_heights(self.coinstate.head().height)
        potential_start_hashes = [self.coinstate.by_height_at_head()[height].hash() for height in heights]
        return GetBlocksMessage(potential_start_hashes)


class RemotePeer:
    def __init__(self, host, port, direction, last_connection_attempt):
        self.host = host
        self.port = port
        self.direction = direction

        self.last_connection_attempt = last_connection_attempt


class DisconnectedRemotePeer(RemotePeer):
    def __init__(self, host, port, direction, last_connection_attempt):
        super().__init__(host, port, direction, last_connection_attempt)

        # self.last_seen_alive = None

    def is_time_to_connect(self, current_time):
        return ((self.last_connection_attempt is None) or
                (current_time - self.last_connection_attempt >= TIME_BETWEEN_CONNECTION_ATTEMPTS))

    def as_connected(self, sock):
        return ConnectedRemotePeer(self.host, self.port, self.direction, self.last_connection_attempt, sock)


class MessageReceiver:
    def __init__(self, peer):
        self.peer = peer

        self.buffer = b''
        self.magic_read = False
        self.len = None

    def receive(self, data):
        self.buffer += data

        if not self.magic_read and len(self.buffer) >= 4:
            magic = self.buffer[:4]
            if magic != MAGIC:
                raise Exception("Insufficient magic")
            else:
                self.magic_read = True

            self.buffer = self.buffer[4:]

        if self.len is None and len(self.buffer) >= 4:
            (self.len,) = struct.unpack(b">I", self.buffer[:4])

            if self.len > MAX_MESSAGE_SIZE:
                raise Exception("len > MAX_MESSAGE_SIZE")

            self.buffer = self.buffer[4:]

        if self.len is not None and self.len <= len(self.buffer):
            self.handle_message_data(self.buffer[:self.len])

            self.buffer = self.buffer[self.len:]
            self.magic_read = False
            self.len = None
            self.receive(b"")  # recurse to repeat (multiple messages could be received in a single socket read)

    def handle_message_data(self, message_data):
        f = BytesIO(message_data)
        header = MessageHeader.stream_deserialize(f)
        message = Message.stream_deserialize(f)
        self.peer.handle_message_received(header, message)


class InventoryMessageState:
    def __init__(self, header, message):
        self.header = header
        self.message = message
        self.index = 0
        self.actually_used = False


class ConnectedRemotePeer(RemotePeer):
    def __init__(self, host, port, direction, last_connection_attempt, sock):
        super().__init__(host, port, direction, last_connection_attempt)
        self.sock = sock
        self.direction = direction

        self.receiver = MessageReceiver(self)
        self.send_backlog = []
        self.send_buffer = b''

        self.hello_sent = False
        self.hello_received = False

        self.sent = b''
        self.received = b''

        self.waiting_for_inventory = False
        self.last_empty_inventory_response_at = 0
        self.inventory_messages = []

        self._next_msg_id = 0
        self.last_get_peers_sent_at = None
        self.waiting_for_peers = False

    def as_disconnected(self):
        return DisconnectedRemotePeer(self.host, self.port, self.direction, self.last_connection_attempt)

    def step(self, current_time):
        """The responsibility of this method: to send the HelloMessage and GetPeersMessage."""

        if not self.hello_sent:
            ipv4_mapped = IPv6Address("::FFFF:%s" % self.host)
            port_if_known = self.port if self.port is not IRRELEVANT else 0

            my_ip_address = IPv6Address("0::0")  # Unspecified
            my_port = local_peer.port

            hello_message = HelloMessage(
                [SupportedVersion(0)], ipv4_mapped, port_if_known, my_ip_address, my_port, local_peer.nonce,
                b"sashimi " + __version__.encode("utf-8"))

            self.hello_sent = True
            self.send_message(hello_message)

        if not self.hello_received:
            return

        if ((self.last_get_peers_sent_at is None or current_time > self.last_get_peers_sent_at + GET_PEERS_INTERVAL)
                and not self.waiting_for_peers):

            self.send_message(GetPeersMessage())
            self.waiting_for_peers = True
            self.last_get_peers_sent_at = current_time

    def _get_msg_id(self):
        self._next_msg_id += 1
        return self._next_msg_id  # 1 is the first message id (0 being reserved for "unknown"). Dijkstra's dead.

    def send_message(self, message, prev_header=None):
        if prev_header is None:
            in_response_to, context = 0, _new_context()
        else:
            in_response_to, context = prev_header.id, prev_header.context
        header = MessageHeader(int(time()), self._get_msg_id(), in_response_to=in_response_to, context=context)

        logger.info(
            "%15s ConnectedRemotePeer.send_message(%s %s)" % (self.host, type(message).__name__, header.format()))

        data = header.serialize() + message.serialize()
        self.send_backlog.append((MAGIC + struct.pack(b">I", len(data)) + data))

        self.check_message_backlog()

    def check_message_backlog(self):
        if len(self.send_buffer) == 0 and len(self.send_backlog) > 0:
            self.send_buffer = self.send_backlog.pop(0)
            self.start_sending()

    def start_sending(self):
        local_peer.selector.modify(self.sock, selectors.EVENT_READ | selectors.EVENT_WRITE, data=self)

    def stop_sending(self):
        local_peer.selector.modify(self.sock, selectors.EVENT_READ, data=self)

    def handle_message_received(self, header, message):
        logger.info("%15s ConnectedRemotePeer.handle_message_received(%s %s)" % (
            self.host, type(message).__name__, header.format()))

        if isinstance(message, HelloMessage):
            return self.handle_hello_message_received(header, message)

        if not self.hello_received:
            raise Exception("First message must be Hello")

        if isinstance(message, GetBlocksMessage):
            return self.handle_get_blocks_message_received(header, message)

        if isinstance(message, InventoryMessage):
            return self.handle_inventory_message_received(header, message)

        if isinstance(message, GetDataMessage):
            return self.handle_get_data_message_received(header, message)

        if isinstance(message, DataMessage):
            return self.handle_data_message_received(header, message)

        if isinstance(message, GetPeersMessage):
            return self.handle_get_peers_message_received(header, message)

        if isinstance(message, PeersMessage):
            return self.handle_peers_message_received(header, message)

        raise NotImplementedError("%s" % message)

    def handle_can_send(self, sock):
        # logger.info("ConnectedRemotePeer.handle_can_send()")

        sent = sock.send(self.send_buffer)
        self.sent += self.send_buffer[:sent]
        self.send_buffer = self.send_buffer[sent:]

        if len(self.send_buffer) == 0:
            self.stop_sending()  # in principle: waisteful, but easy to reason about.
            self.check_message_backlog()

    def handle_receive_data(self, data):
        # logger.info("%15s ConnectedRemotePeer.handle_receive_data()" % self.host)
        self.received += data

        self.receiver.receive(data)

    def handle_hello_message_received(self, header, message):
        logger.info("%15s ConnectedRemotePeer.handle_hello_message_received(%s)" % (self.host, message.user_agent))
        self.hello_received = True

        if self.direction == INCOMING:
            # also add the peer to the list of disconnected_peers in reverse direction
            # TODO at some point: this means that if possible, each peer will be connected to twice. We should drop
            # connection if that happens. Probably like so: it's the prober that receives "hello" and then closes
            # because satisifed.
            key = (self.host, message.my_port, OUTGOING)
            nm = local_peer.network_manager
            nm._sanity_check()
            if key not in nm.disconnected_peers and key not in nm.connected_peers:
                nm.disconnected_peers[key] = DisconnectedRemotePeer(self.host, message.my_port, OUTGOING, None)
            nm._sanity_check()

        if self.direction == OUTGOING and message.nonce == local_peer.nonce:
            local_peer.network_manager.my_addresses.add((self.host, self.port))
            local_peer.disconnect(self, "connection to self")

        local_peer.network_manager.update_peer_db(self)

    def handle_get_blocks_message_received(self, header, message):
        logger.info("%15s ConnectedRemotePeer.handle_get_blocks_message_received()" % self.host)
        coinstate = local_peer.chain_manager.coinstate
        logger.debug("%15s ... at coinstate %s" % (self.host, coinstate))
        for potential_start_hash in message.potential_start_hashes:
            logger.debug("%15s ... psh %s" % (self.host, human(potential_start_hash)))
            if potential_start_hash in coinstate.block_by_hash:
                start_height = coinstate.block_by_hash[potential_start_hash].height + 1  # + 1: sent hash is last known
                if start_height not in coinstate.by_height_at_head():
                    # we have no new info
                    logger.debug("%15s ... no new info" % self.host)
                    self.send_message(InventoryMessage([]), prev_header=header)
                    return

                if coinstate.by_height_at_head()[start_height].previous_block_hash == potential_start_hash:
                    # this final if checks that this particular potential_start_hash is on our active chain
                    break
        else:  # no break
            start_height = 1  # genesis is last known

        max_height = coinstate.head().height + 1  # + 1: range is exclusive, but we need to send this last block also
        items = [
            InventoryItem(DATA_BLOCK, coinstate.by_height_at_head()[height].hash())
            for height in range(start_height, min(start_height + GET_BLOCKS_INVENTORY_SIZE, max_height))
        ]
        logger.debug("%15s ... returning from %s, %s items" % (self.host, start_height, len(items)))
        self.send_message(InventoryMessage(items), prev_header=header)

    def handle_inventory_message_received(self, header, message):
        logger.info("%15s ConnectedRemotePeer.handle_inventory_message_received(%s)" % (self.host, len(message.items)))
        if len(message.items) > 0:
            logger.info("%15s %s .. %s" % (self.host, human(message.items[0].hash), human(message.items[-1].hash)))

        if len(message.items) > GET_BLOCKS_INVENTORY_SIZE:
            raise Exception("Inventory msg too big")

        self.waiting_for_inventory = False

        if message.items == []:
            logger.info("%15s ConnectedRemotePeer.last_empty_inventory_response_at set" % self.host)
            self.last_empty_inventory_response_at = int(time())  # TODO time() as a pass-along?
            return

        self.inventory_messages.append(InventoryMessageState(header, message))
        self.check_inventory_messages()

    def _get_hash_from_inventory_messages(self):
        while self.inventory_messages:
            msg_state = self.inventory_messages[0]

            if msg_state.index <= len(msg_state.message.items) - 1:
                if msg_state.message.items[msg_state.index].data_type != DATA_BLOCK:
                    raise Exception("We only deal w/ Block InventoryMessage for now")

                result = msg_state.message.items[msg_state.index].hash
                msg_state.index += 1
                return msg_state, result

            if not msg_state.actually_used and len(msg_state.message.items) > 0:
                logger.info("%15s ConnectedRemotePeer._get_hash_from_inventory_messages: try again from %s" % (
                    self.host, human(msg_state.message.items[-1].hash)))

                get_blocks_message = GetBlocksMessage([msg_state.message.items[-1].hash])
                self.waiting_for_inventory = True
                # TODO consider adding something like the below
                # self.actively_fetching_blocks_from_peers.append((current_time + IBD_PEER_TIMEOUT, remote_peer))
                self.send_message(get_blocks_message, prev_header=msg_state.header)

            self.inventory_messages.pop(0)

        return None, None

    def check_inventory_messages(self):
        coinstate = local_peer.chain_manager.coinstate

        msg_state, next_hash = self._get_hash_from_inventory_messages()
        while next_hash is not None and next_hash in coinstate.block_by_hash:
            msg_state, next_hash = self._get_hash_from_inventory_messages()

        if next_hash is None or next_hash in coinstate.block_by_hash:
            return

        msg_state.actually_used = True
        self.send_message(GetDataMessage(DATA_BLOCK, next_hash), prev_header=msg_state.header)

    def handle_get_data_message_received(self, header, get_data_message):
        if get_data_message.data_type != DATA_BLOCK:
            raise NotImplementedError("We can only deal w/ DATA_BLOCK GetDataMessage objects for now")

        coinstate = local_peer.chain_manager.coinstate
        if get_data_message.hash not in coinstate.block_by_hash:
            # we simply silently ignore GetDataMessage for hashes we don't have... future work: inc banscore, or ...
            logger.debug("%15s ConnectedRemotePeer.handle_data_message_received for unknown hash %s" % (
                self.host, human(get_data_message.hash)))
            return

        data_message = DataMessage(DATA_BLOCK, coinstate.block_by_hash[get_data_message.hash])

        logger.debug("%15s ConnectedRemotePeer.handle_data_message_received for hash %s h. %s" % (
            self.host, human(get_data_message.hash), coinstate.block_by_hash[get_data_message.hash].height))
        self.send_message(data_message, prev_header=header)

    def handle_data_message_received(self, header, message):
        logger.info("%15s ConnectedRemotePeer.handle_data_message_received(%s %s)" % (
            self.host, message.data_type, header.format()))

        if message.data_type == DATA_BLOCK:
            return self.handle_block_received(header, message)

        if message.data_type == DATA_TRANSACTION:
            return self.handle_transaction_received(header, message)

        raise NotImplementedError("Unknown DataMessage objects for now")

    def handle_block_received(self, header, message):
        # TODO deal with out-of-order blocks more gracefully

        block = message.data
        coinstate = local_peer.chain_manager.coinstate

        if block.hash() in coinstate.block_by_hash:
            # implicit here: when you receive a datamessage, this could be because you requested it; and thus you'll
            # want to request the next one (the below check does nothing if there's nothing to do)
            self.check_inventory_messages()

            logger.debug("%15s ConnectedRemotePeer.handle_block_received() for known block %s" % (
                self.host, human(block.hash())))

            # Known block, just return
            return

        try:
            coinstate = coinstate.add_block(block, int(time()))  # TODO time() as a pass-along?
            local_peer.chain_manager.set_coinstate(coinstate)

            # TODO writing blocks to disk should probably not occur here, and should probably a bit more sophisticated.
            with open('chain/%s' % block_filename(block), 'wb') as f:
                f.write(block.serialize())
        except Exception as e:
            # TODO: dirty hack at this particular point... to allow for out-of-order blocks to not take down the whole
            # peer, but this should more specifically match for a short list of OK problems.
            logger.info("%15s INVALID block %s" % (self.host, str(e)))

        # implicit here: when you receive a datamessage, this is because you requested it; and thus you'll want to
        # request the next one
        self.check_inventory_messages()

        if block == coinstate.head() and header.in_response_to == 0:
            # "header.in_response_to == 0" is being used as a bit of a proxy for "not in IBD" here, but it would be
            # better to check for that state more explicitly. We don't want to broadcast blocks while in IBD, because in
            # that state the fact that some block is our new head doesn't mean at all that we're talking about the real
            # chain's new head, and only the latter is relevant to the rest of the world.
            local_peer.network_manager.broadcast_block(block)

    def handle_transaction_received(self, header, message):
        transaction = message.data
        if transaction in local_peer.chain_manager.transaction_pool:
            return

        if local_peer.chain_manager.add_transaction_to_pool(transaction):
            # if this is valid and new... just broadcast it to every peer you know. I'm sure this is inefficient, but
            # at least peers will stop broadcasting once they receive it a second time themselves.
            local_peer.network_manager.broadcast_transaction(transaction)

    def handle_get_peers_message_received(self, header, message):
        peers = []

        for peer in local_peer.network_manager.connected_peers.values():
            if peer.direction == OUTGOING:
                peers.append(Peer(int(time()), IPv6Address("::FFFF:%s" % peer.host), peer.port))

        for peer in local_peer.network_manager.disconnected_peers.values():
            if peer.direction == OUTGOING:
                peers.append(Peer(0, IPv6Address("::FFFF:%s" % peer.host), peer.port))  # TODO 'last seen' time.

        # TODO filter out local network addresses (also on the receiving end)
        peers = peers[:1000]  # send 1000 peers max.

        self.send_message(PeersMessage(peers), prev_header=header)

    def handle_peers_message_received(self, header, message):
        # TODO peers that have been communicated to you like this should be marked as "not checked yet" somehow, to
        # avoid being flooded with nonsense peers.

        self.waiting_for_peers = False

        for announced_peer in message.peers:
            ipv4_mapped = announced_peer.ip_address.ipv4_mapped
            if ipv4_mapped is None:
                continue  # IPv6? Ain't nobody got time for that! (Seriously though, the protocol supports it if needed)
            host = ipv4_mapped.exploded

            # TODO factor out copypasta
            key = (host, announced_peer.port, OUTGOING)
            nm = local_peer.network_manager
            nm._sanity_check()
            if key not in nm.disconnected_peers and key not in nm.connected_peers:
                nm.disconnected_peers[key] = DisconnectedRemotePeer(host, announced_peer.port, OUTGOING, None)
            nm._sanity_check()


class LocalPeer:

    def __init__(self):
        self.port = None  # perhaps just push this into the signature here?
        self.nonce = random.randrange(pow(2, 32))
        self.selector = selectors.DefaultSelector()
        self.network_manager = NetworkManager()
        self.chain_manager = ChainManager(int(time()))
        self.managers = [
            self.network_manager,
            self.chain_manager,
        ]

    def start_listening(self, port=PORT):
        self.port = port
        logger.info("%15s LocalPeer.start_listening(%s)" % ("", port))
        lsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # https://stackoverflow.com/questions/4465959/python-errno-98-address-already-in-use/4466035#4466035
        lsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        lsock.bind(("", port))
        lsock.listen()
        lsock.setblocking(False)
        self.selector.register(lsock, selectors.EVENT_READ, data=LISTENING_SOCKET)

    def handle_incoming_connection(self, sock):
        logger.info("%15s LocalPeer.handle_incoming_connection()" % "")
        # TODO only accept a single (incoming, outgoing) connection from each peer
        conn, addr = sock.accept()
        conn.setblocking(False)
        events = selectors.EVENT_READ

        remote_host = conn.getpeername()[0]
        remote_port = conn.getpeername()[1]
        remote_peer = ConnectedRemotePeer(remote_host, remote_port, INCOMING, None, conn)
        self.selector.register(conn, events, data=remote_peer)
        self.network_manager.handle_peer_connected(remote_peer)

    def handle_remote_peer_selector_event(self, key, mask):
        # logger.info("LocalPeer.handle_remote_peer_selector_event()")

        sock = key.fileobj
        remote_peer = key.data
        assert isinstance(remote_peer, ConnectedRemotePeer)

        try:
            if mask & selectors.EVENT_READ:
                recv_data = sock.recv(1024)

                if recv_data:
                    remote_peer.handle_receive_data(recv_data)
                else:
                    self.disconnect(remote_peer, "connection closed remotely")  # is this so?

            if mask & selectors.EVENT_WRITE:
                remote_peer.handle_can_send(sock)

        except OSError as e:  # e.g. ConnectionRefusedError, "Bad file descriptor"
            # no print-to-screen for this one
            logger.info("%15s Disconnecting remote peer %s" % (remote_peer.host, e))
            self.disconnect(remote_peer, "OS error")

        except ValueError as e:  # e.g. Invalid file descriptor: {}".format(fd)) ... more exact matching is better tho
            # no print-to-screen for this one
            logger.info("%15s Disconnecting remote peer %s" % (remote_peer.host, e))
            self.disconnect(remote_peer, "OS error")

        except Exception as e:
            print(traceback.format_exc())  # be loud... this is likely a programming error.
            # We take the position that any exception caused is reason to disconnect. This allows the code that talks to
            # peers to not have special cases for exceptions since they will all be caught by this catch-all.
            logger.info("%15s Disconnecting remote peer %s" % (remote_peer.host, e))
            self.disconnect(remote_peer, "Exception")

    def disconnect(self, remote_peer, reason=""):
        logger.info("%15s LocalPeer.disconnect(%s)" % (remote_peer.host, reason))

        try:
            self.selector.unregister(remote_peer.sock)
            remote_peer.sock.close()
            self.network_manager.handle_peer_disconnected(remote_peer.as_disconnected())
        except Exception as e:
            # yes yes... sweeping things under the carpet here. until I actually RTFM and think this through
            # (i.e. the whole business of unregistering things that are already in some half-baked state)
            # at least one path how you might end up here: a EVENT_WRITE is reached for a socket that was just closed
            # as a consequence of something that was read.
            logger.info("%15s Error while disconnecting %s" % ("", e))

    def start_outgoing_connection(self, disconnected_peer):
        logger.info("%15s LocalPeer.start_outgoing_connection()" % disconnected_peer.host)

        server_addr = (disconnected_peer.host, disconnected_peer.port)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setblocking(False)
        sock.connect_ex(server_addr)
        events = selectors.EVENT_READ

        remote_peer = disconnected_peer.as_connected(sock)
        self.selector.register(sock, events, data=remote_peer)
        self.network_manager.handle_peer_connected(remote_peer)

    def step_managers(self, current_time):
        for manager in self.managers:
            manager.step(current_time)

    def handle_selector_events(self):
        events = self.selector.select(timeout=1)  # TODO this is for the managers to do something... tune it though
        for key, mask in events:
            if key.data is LISTENING_SOCKET:
                self.handle_incoming_connection(key.fileobj)
            else:
                self.handle_remote_peer_selector_event(key, mask)

    def run(self):
        self.running = True
        try:
            while self.running:
                current_time = int(time())
                self.step_managers(current_time)
                self.handle_selector_events()
        except KeyboardInterrupt:
            print("caught keyboard interrupt, exiting")
        finally:
            self.selector.close()

    def stop(self):
        self.running = False

    def show_stats(self):
        coinstate = self.chain_manager.coinstate

        print("NETWORK")
        print("Nr. of connected peers:", len(self.network_manager.get_active_peers()))
        for p in self.network_manager.get_active_peers()[:10]:
            print("%15s:%s - %s" % (p.host, p.port if p.port != IRRELEVANT else "....", p.direction))

        print("\nCHAIN")
        for (head, lca) in coinstate.forks():
            if head.height < coinstate.head().height - 10:
                continue  # don't show forks which are out-ran by more than 10 blocks

            print("Height    %s" % head.height)
            print("Date/time %s" % datetime.fromtimestamp(head.timestamp).isoformat())
            if head.height != lca.height:
                print("diverges for %s blocks" % (head.height - lca.height))
            print()


local_peer = LocalPeer()



from .humans import human
from .hash import sha256d


def get_merkle_root(list_of_hashes):
    if len(list_of_hashes) == 1:
        return list_of_hashes[0]

    new_list = []
    for chunk in _chunks(list_of_hashes, 2):
        if len(chunk) == 2:
            new_list.append(sha256d(chunk[0] + chunk[1]))
        else:  # implied: len(chunk) == 1
            new_list.append(chunk[0])

    return get_merkle_root(new_list)


class MerkleNode:
    def __init__(self, index, children, value=None):
        # for non-leaves index is a lower-bound of all leaves' indexes
        self.index = index
        self.children = children
        self.value = value

    def hash(self):
        if len(self.children) > 0 and self.value is not None:
            raise ValueError("Only store values at the leaves")

        if self.value is not None:
            return self.value

        return sha256d(b''.join(c.hash() for c in self.children))

    def __repr__(self):
        return "M(%s, (%s))" % (human(self.hash())[:7], self.children)


def _get_merkle_tree(list_of_nodes):
    if len(list_of_nodes) == 1:
        return list_of_nodes[0]

    new_list = []
    for chunk in _chunks(list_of_nodes, 2):
        if len(chunk) == 2:
            new_list.append(MerkleNode(chunk[0].index, (chunk[0], chunk[1])))
        else:  # implied: len(chunk) == 1
            new_list.append(chunk[0])

    return _get_merkle_tree(new_list)


def get_merkle_tree(list_of_hashes):
    return _get_merkle_tree([MerkleNode(i, (), h) for (i, h) in enumerate(list_of_hashes)])


def get_proof(merkle_node, index_of_interest):
    if not merkle_node.children:
        return merkle_node

    # our nodes always have either 0 or 2 children, never 1
    if index_of_interest >= merkle_node.children[1].index:
        other, recurse_into = merkle_node.children
        reconstruct = lambda ot, rec: (ot, rec)  # noqa
    else:
        recurse_into, other = merkle_node.children
        reconstruct = lambda ot, rec: (rec, ot) # noqa

    simplified_other = MerkleNode(other.index, (), other.hash())
    recursion_result = get_proof(recurse_into, index_of_interest)

    return MerkleNode(merkle_node.index, reconstruct(simplified_other, recursion_result))


def _chunks(l, chunk_size):
    """return chunks of chunk_size for list l

    >>> list(chunks([0, 1, 2, 3, 4], 2))
    [[0, 1], [2, 3], [4]]
    """
    return (l[i:i + chunk_size] for i in range(0, len(l), chunk_size))



from copy import deepcopy
import immutables
import pytest

from skepticoin.humans import computer
from skepticoin.params import SASHIMI_PER_COIN, MAX_COINBASE_RANDOM_DATA_SIZE
from skepticoin.coinstate import CoinState
from skepticoin.consensus import (
    construct_minable_summary_genesis,
    construct_minable_summary,
    construct_coinbase_transaction,
    construct_pow_evidence,
    get_block_subsidy,
    get_transaction_fee,
    validate_coinbase_transaction_by_itself,
    validate_block_header_by_itself,
    validate_block_by_itself,
    # validate_non_coinbase_transaction_in_coinstate,
    ValidateBlockError,
    ValidateBlockHeaderError,
    ValidateTransactionError,
    ValidatePOWError,
)
from skepticoin.signing import SECP256k1PublicKey, SECP256k1Signature
from skepticoin.datatypes import Transaction, OutputReference, Input, Output, Block


example_public_key = SECP256k1PublicKey(b'x' * 64)


def get_example_genesis_block():
    return Block.deserialize(computer(
        """000000000000000000000000000000000000000000000000000000000000000000008278968af4bd613aa24a5ccd5280211b3101e3"""
        """ff62621bb11500509d1bbe2a956046240b0100000000000000000000000000000000000000000000000000000000000000000000d7"""
        """38f2c472180cb401f650b12be96ec25bfd9b4e9908c6c9089d9bf26401646f87000000000000000000000000000000000000000000"""
        """0000000000000000000000077a14cfbe21d47f367f23f9a464c765541b1b07bef9f5a95901e0bffe3a1a2f01000100000000000000"""
        """000000000000000000000000000000000000000000000000000000000001000000000001000000012a05f200027878787878787878"""
        """7878787878787878787878787878787878787878787878787878787878787878787878787878787878787878787878787878787878"""
        """787878"""))


def test_construct_minable_summary_no_transactions():
    summary = construct_minable_summary_genesis([
        construct_coinbase_transaction(0, [], immutables.Map(), b"Political statement goes here", example_public_key)
    ], 1231006505, 0)

    summary.serialize()
    # so far... just checking that this doesn't crash :-)


def test_construct_minable_summary_with_transactions():
    # TODO
    '''
    coinstate = CoinState.empty()
    construct_minable_summary(coinstate, [
        construct_coinbase_transaction(0, [], immutables.Map(), b"Political statement goes here", example_public_key)
    ], 1231006505, 0)
    '''


def test_construct_pow_evidence_genesis_block():
    coinstate = CoinState.empty()

    transactions = [
        construct_coinbase_transaction(0, [], immutables.Map(), b"Political statement goes here", example_public_key),
    ]

    summary = construct_minable_summary_genesis(transactions, 1231006505, 0)

    evidence = construct_pow_evidence(coinstate, summary, 0, transactions)
    evidence.serialize()
    # no assertions here, just checking that this doesn't crash :-)


def test_construct_pow_evidence_blahblahblabl():
    # TODO
    pass


def test_validate_non_coinbase_transaction_by_itself_no_inputs():
    pass  # TODO


def test_validate_non_coinbase_transaction_by_itself_no_outputs():
    pass  # TODO


def test_validate_non_coinbase_transaction_by_itself_max_size():
    pass  # TODO


def test_validate_non_coinbase_transaction_by_itself_max_total_output():
    pass  # TODO


def test_validate_non_coinbase_transaction_by_itself_no_duplicate_transactions():
    pass  # TODO


def test_validate_non_coinbase_transaction_by_itself_is_not_coinbase():
    pass  # TODO


def test_validate_signature_for_spend():
    # before anything: create sign_for_spend
    pass  # TODO


def test_get_transaction_fee():
    public_key = SECP256k1PublicKey(b'x' * 64)

    previous_transaction_hash = b'a' * 32

    unspent_transaction_outs = immutables.Map({
        OutputReference(previous_transaction_hash, 0): Output(40, public_key),
        OutputReference(previous_transaction_hash, 1): Output(34, public_key),
    })

    transaction = Transaction(
        inputs=[Input(
            OutputReference(previous_transaction_hash, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(30, public_key)]
    )

    assert get_transaction_fee(transaction, unspent_transaction_outs) == 34 - 30  # 4


def test_get_block_fees():
    pass  # TODO make a test for this once no longer trivial (i.e. when spending from the same block is allowed)


def test_get_block_subsidy():
    assert get_block_subsidy(0) == 10 * SASHIMI_PER_COIN
    assert get_block_subsidy(1_049_999) == 10 * SASHIMI_PER_COIN
    assert get_block_subsidy(1_050_000) == 5 * SASHIMI_PER_COIN
    assert get_block_subsidy(31_499_999) == 1
    assert get_block_subsidy(31_500_000) == 0


def _get_example_coinbase_transaction():
    height = 123
    non_coinbase_transactions = []
    unspent_transaction_outs = immutables.Map()
    miner_public_key = b'x' * 32
    random_data = b"No need to get all political here"
    return construct_coinbase_transaction(
        height, non_coinbase_transactions, unspent_transaction_outs, random_data, miner_public_key)


def test_validate_coinbase_transaction_by_itself_for_valid_coinbase():
    cb = _get_example_coinbase_transaction()
    validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_exactly_1_input():
    cb = _get_example_coinbase_transaction()
    cb.inputs.append(deepcopy(cb.inputs[0]))

    with pytest.raises(ValidateTransactionError, match=".*1 input.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_reference_real_output():
    cb = _get_example_coinbase_transaction()
    cb.inputs[0].output_reference = OutputReference(b'c' * 32, 4)

    with pytest.raises(ValidateTransactionError, match=".*must create.*thin air.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_should_have_coinbasedata():
    cb = _get_example_coinbase_transaction()
    cb.inputs[0].signature = SECP256k1Signature(b'c' * 64)

    with pytest.raises(ValidateTransactionError, match=".*CoinbaseData.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_coinbase_transaction_by_itself_maximum_coinbasedata_size():
    cb = _get_example_coinbase_transaction()
    cb.inputs[0].signature.signature = b'x' * (MAX_COINBASE_RANDOM_DATA_SIZE + 1)

    with pytest.raises(ValidateTransactionError, match=".*MAX_COINBASE_RANDOM_DATA_SIZE.*"):
        validate_coinbase_transaction_by_itself(cb)


def test_validate_block_header_by_itself_for_correct_header():
    validate_block_header_by_itself(get_example_genesis_block().header, 1615209942)


def test_validate_block_header_by_itself_for_bad_pow():
    block = get_example_genesis_block()
    block.header.summary.nonce = 1
    with pytest.raises(ValidatePOWError):
        validate_block_header_by_itself(block.header, 1615209942)


def test_validate_block_header_by_itself_for_block_from_the_future():
    block = get_example_genesis_block()
    with pytest.raises(ValidateBlockHeaderError):
        validate_block_header_by_itself(block.header, block.header.summary.timestamp - 100)


def test_validate_block_by_itself_for_correct_block():
    validate_block_by_itself(get_example_genesis_block(), 1615209942)


def test_validate_block_by_itself():
    # TODO we haven't tested actual failures yet.
    pass  # TODO  should probably be split into multiple tests


def test_validate_block_by_itself_for_mismatched_heights():
    block = get_example_genesis_block()
    block.height = 1

    with pytest.raises(ValidateBlockError, match=".*height.*"):
        validate_block_by_itself(block, 1615209942)


def test_validate_non_coinbase_transaction_in_coinstate_invalid_output_reference():
    # I started on the below, but the amount of setup is getting excessive... perhaps it's going to be easier to
    # express these tests when more mechanisms of _creation_ are available? we'll see
    '''
    previous_transaction_hash = b'a' * 32

    unspent_transaction_outs = immutables.Map()

    transaction = Transaction(
        inputs=[Input(
            OutputReference(previous_transaction_hash, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[Output(30, public_key)]
    )

    with pytest.raises(ValidateTransactionError, match=r".*does not exist.*") as e:
        validate_non_coinbase_transaction_in_coinstate

    '''


def test_validate_non_coinbase_transaction_in_coinstate_invalid_signature():
    pass  # TODO


def test_validate_non_coinbase_transaction_in_coinstate_overspending():
    pass  # TODO


def test_validate_non_coinbase_transaction_in_coinstate_valid():
    pass  # TODO



from skepticoin.signing import (
    PublicKey, SignableEquivalent, CoinbaseData, Signature, SECP256k1Signature, SECP256k1PublicKey)


def serialize_and_deserialize(thing, clz):
    other_thing = clz.deserialize(thing.serialize())
    assert thing == other_thing


def test_signable_equivalent_serialization():
    se = SignableEquivalent()
    serialize_and_deserialize(se, SignableEquivalent)


def test_coinbase_data_serialization():
    rd = CoinbaseData(1235, b"\x19\x88\x55")
    serialize_and_deserialize(rd, Signature)
    assert repr(rd) == 'CoinbaseData(1235, #198855)'


def test_coinbase_data_pretty_repr():
    rd = CoinbaseData(188, b"Don't trust the government, trust me instead.")
    assert repr(rd) == 'CoinbaseData(188, "Don\'t trust the government, trust me instead.")'


def test_signature_serialization():
    sig = SECP256k1Signature(b"a" * 64)
    serialize_and_deserialize(sig, Signature)


def test_publickey_serialization():
    pk = SECP256k1PublicKey(b"5" * 64)
    serialize_and_deserialize(pk, PublicKey)



SASHIMI_PER_COIN = 100_000_000  # bitcoin calls this (perhaps somewhat too) simply "COIN"

# As is customary in the alt-coin sphere, we mess with the Bitcoin params a bit in order to generate an air of
# uniqueness. Ahum, I meant to say that the following parameters have been carefully chosen to match Bitcoin's total
# supply of money (remember: 1 skepticoin == 1 btc). However, we have made the block duration significantly shorter to
# provide a more interactive experience.

FIVE = (10 // 2)  # every 2 minutes. bitcoin has this at 10, so we change the params below with a factor of FIVE
DESIRED_BLOCK_TIMESPAN = 10 * 60 // FIVE
BLOCKS_BETWEEN_TARGET_READJUSTMENT = 2016 * FIVE
DESIRED_TARGET_READJUSTMENT_TIMESPAN = BLOCKS_BETWEEN_TARGET_READJUSTMENT * DESIRED_BLOCK_TIMESPAN  # every 2 weeks

INITIAL_SUBSIDY = 50 * SASHIMI_PER_COIN // FIVE
SUBSIDY_HALVING_INTERVAL = 210_000 * FIVE  # hype train departs every 4 years

MAX_BLOCK_SIZE = 1_000_000 // FIVE
MAX_COINBASE_RANDOM_DATA_SIZE = 200

# The total amount of money can be deduced from the halving algorithm but is put here as a constant; for bitcoin it is
# often said to be "21 million" but is actually 2_099_999_997_690_000 satoshi. Because of rounding differences there
# will be slightly fewer sashimi. The only logical conclusion is that skepticoin is slightly more valuable (deflation!)
# Note that this fits in 48 bits (and therefore easily in 8 bytes).
MAX_SASHIMI = 2_099_999_986_350_000

# this is the awesome script to determine this value:
"""
>>> total_coin = 0
... subsidy = 10 * 100_000_000
... while subsidy > 0:
...     total_coin += 5 * 210_000 * subsidy
...     subsidy = subsidy // 2
...     print(total_coin, subsidy)
"""

# We start with an arbitrarily low target; recalibration will follow soon anyway.
INITIAL_TARGET = (1 << (8 * 31)).to_bytes(length=32, byteorder='big')

# bitcoin, and many of its clones, actually allow for blocks to be mined up to 2 hours into the future. The reason given
# is that it's impossible to synchronize clocks in distributed networks. True as that may be, surely it's not hard to
# synchronize them "mostly correctly"? i.e. in the couple-of-seconds range? The constant below should be plenty.
MAX_FUTURE_BLOCK_TIME = 30


# POW Evidence
CHAIN_SAMPLE_COUNT = 8
CHAIN_SAMPLE_SIZE = 4
CHAIN_SAMPLE_TOTAL_SIZE = CHAIN_SAMPLE_COUNT * CHAIN_SAMPLE_SIZE



from skepticoin.pow import select_block_height, select_block_slice
from skepticoin.hash import sha256d


list_of_hashes = [sha256d(b'x' * i) for i in range(40)]


def test_select_block_height():
    for hash in list_of_hashes:
        assert select_block_height(hash, 1) in [0]
        assert select_block_height(hash, 2) in [0, 1]


def test_select_block_slice():
    assert b'a really short blocka r' == select_block_slice(b'xxxxxxxx\00\00\00\00', b'a really short block', 23)

    assert b'really short blocka rea' == select_block_slice(b'xxxxxxxx\00\00\00\02', b'a really short block', 23)

    assert b'short blocka really sho' == select_block_slice(b'xxxxxxxx\f0\29\00\02', b'a really short block', 23)



import immutables

from skepticoin.signing import CoinbaseData, SECP256k1PublicKey, SECP256k1Signature
from skepticoin.datatypes import Transaction, OutputReference, Input, Output
from skepticoin.coinstate import uto_apply_transaction, pkb_apply_transaction, PKBalance
from skepticoin.consensus import construct_reference_to_thin_air


def test_uto_apply_transaction_on_coinbase():
    public_key = SECP256k1PublicKey(b'x' * 64)

    output_0 = Output(40, public_key)
    output_1 = Output(34, public_key)

    unspent_transaction_outs = immutables.Map()

    transaction = Transaction(
        inputs=[Input(
            construct_reference_to_thin_air(),
            CoinbaseData(0, b'coinbase of the first block'),
        )],
        outputs=[output_0, output_1]
    )

    result = uto_apply_transaction(unspent_transaction_outs, transaction, is_coinbase=True)

    assert OutputReference(transaction.hash(), 0) in result
    assert OutputReference(transaction.hash(), 1) in result

    assert result[OutputReference(transaction.hash(), 0)] == output_0
    assert result[OutputReference(transaction.hash(), 1)] == output_1


def test_uto_apply_transaction_on_non_coinbase_transaction():
    public_key = SECP256k1PublicKey(b'x' * 64)

    output_0 = Output(40, public_key)
    output_1 = Output(34, public_key)
    output_2 = Output(30, public_key)

    previous_transaction_hash = b'a' * 32

    unspent_transaction_outs = immutables.Map({
        OutputReference(previous_transaction_hash, 0): output_0,
        OutputReference(previous_transaction_hash, 1): output_1,
    })

    transaction = Transaction(
        inputs=[Input(
            OutputReference(previous_transaction_hash, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[output_2]
    )

    result = uto_apply_transaction(unspent_transaction_outs, transaction, is_coinbase=False)

    assert OutputReference(previous_transaction_hash, 0) in result
    assert OutputReference(previous_transaction_hash, 1) not in result  # spent
    assert OutputReference(transaction.hash(), 0) in result

    assert result[OutputReference(previous_transaction_hash, 0)] == output_0
    assert result[OutputReference(transaction.hash(), 0)] == output_2


def test_pkb_apply_transaction_on_coinbase():
    public_key_0 = SECP256k1PublicKey(b'0' * 64)
    public_key_1 = SECP256k1PublicKey(b'1' * 64)

    output_0 = Output(40, public_key_0)
    output_1 = Output(34, public_key_1)

    unspent_transaction_outs = immutables.Map()
    public_key_balances = immutables.Map()

    transaction = Transaction(
        inputs=[Input(
            construct_reference_to_thin_air(),
            CoinbaseData(0, b'coinbase of the first block'),
        )],
        outputs=[output_0, output_1]
    )

    result = pkb_apply_transaction(unspent_transaction_outs, public_key_balances, transaction, is_coinbase=True)

    assert public_key_0 in result
    assert public_key_1 in result

    assert result[public_key_0].value == 40
    assert result[public_key_1].value == 34


def test_pkb_apply_transaction_on_non_coinbase_transaction():
    public_key_0 = SECP256k1PublicKey(b'\x00' * 64)
    public_key_1 = SECP256k1PublicKey(b'\x01' * 64)
    public_key_2 = SECP256k1PublicKey(b'\x02' * 64)

    output_0 = Output(40, public_key_0)
    output_1 = Output(34, public_key_1)
    output_3 = Output(66, public_key_1)
    final_output = Output(30, public_key_2)

    previous_transaction_hash = b'a' * 32

    unspent_transaction_outs = immutables.Map({
        OutputReference(previous_transaction_hash, 0): output_0,
        OutputReference(previous_transaction_hash, 1): output_1,
        OutputReference(previous_transaction_hash, 2): output_3,
    })

    public_key_balances = immutables.Map({
        public_key_0: PKBalance(0, []),
        public_key_1: PKBalance(100, [
            OutputReference(previous_transaction_hash, 1),
            OutputReference(previous_transaction_hash, 2),
        ]),
    })

    transaction = Transaction(
        inputs=[Input(
            OutputReference(previous_transaction_hash, 1),
            SECP256k1Signature(b'y' * 64),
        )],
        outputs=[final_output]
    )

    result = pkb_apply_transaction(unspent_transaction_outs, public_key_balances, transaction, is_coinbase=False)

    assert result[public_key_0].value == 0  # not referenced in the transaction under consideration
    assert result[public_key_0].output_references == []

    assert result[public_key_1].value == 100 - 34
    assert result[public_key_1].output_references == [OutputReference(previous_transaction_hash, 2)]

    assert result[public_key_2].value == 30  # the value of the transaction output
    assert result[public_key_2].output_references == [OutputReference(transaction.hash(), 0)]



from io import BytesIO
import struct


class DeserializationError(Exception):
    pass


class SerializationError(Exception):
    pass


class SerializationTruncationError(SerializationError):
    pass


class Serializable:
    def serialize(self):
        f = BytesIO()
        self.stream_serialize(f)
        return f.getvalue()

    @classmethod
    def deserialize(cls, bytes_):
        f = BytesIO(bytes_)
        f.seek(0)
        return cls.stream_deserialize(f)


def safe_read(f, n):
    r = f.read(n)

    if len(r) < n:
        raise SerializationTruncationError('Requested %i bytes but got %i' % (n, len(r)))

    return r


def stream_serialize_list(f, l):
    stream_serialize_vlq(f, len(l))
    for elem in l:
        elem.stream_serialize(f)


def stream_deserialize_list(f, clz):
    result = []
    length = stream_deserialize_vlq(f)
    for i in range(length):
        result.append(clz.stream_deserialize(f))
    return result


def serialize_list(l):
    f = BytesIO()
    stream_serialize_list(f, l)
    return f.getvalue()


def deserialize_list(cls, bytes_):
    f = BytesIO(bytes_)
    f.seek(0)
    return cls.stream_deserialize(f)


def stream_serialize_vlq(f, i):
    r"""From wikipedia: https://en.wikipedia.org/wiki/Variable-length_quantity

    The encoding assumes an octet (an eight-bit byte) where the most significant bit (MSB), also commonly known as the
    sign bit, is reserved to indicate whether another VLQ octet follows.

    If the MSB is 0, then this is the last VLQ octet of the integer, if it is 1, then another VLQ octet follows. The
    other 7 bits are treated as a 7-bit number. The VLQ octets are arranged most significant first in a stream."""
    needed_bytes = (i.bit_length() // 7) + 1

    mod = None
    for j in reversed(range(needed_bytes)):
        div = pow(128, j)
        f.write(struct.pack(b"B", (i % mod if mod else i) // div + (128 if j > 0 else 0)))
        mod = div


def stream_deserialize_vlq(f):
    """ """
    result = 0

    while True:
        (b,) = struct.unpack(b"B", safe_read(f, 1))

        result += (b % 128)

        if b < 128:
            return result

        result *= 128



import threading

from .peer import local_peer
from .params import PORT


class NetworkingThread(threading.Thread):

    def __init__(self, coinstate, port=PORT):
        super().__init__(name="NetworkingThread")
        self.port = port

        self.local_peer = local_peer
        local_peer.chain_manager.set_coinstate(coinstate)

    def run(self):
        self.local_peer.start_listening(self.port)
        self.local_peer.run()

    def stop(self):
        self.local_peer.stop()



from skepticoin.merkletree import get_merkle_root, get_merkle_tree, get_proof
from skepticoin.hash import sha256d


list_of_hashes = [sha256d(b'x' * i) for i in range(40)]


def test_compare_2_ways_of_calculating_root():
    # truncate our example data at every non-zero length to get a nice test set
    for truncate in range(1, len(list_of_hashes)):
        assert get_merkle_tree(list_of_hashes[:truncate]).hash() == get_merkle_root(list_of_hashes[:truncate])


def test_get_proof():
    for truncate in range(1, len(list_of_hashes)):
        tree = get_merkle_tree(list_of_hashes[:truncate])
        for index_of_interest in range(truncate):
            proof = get_proof(tree, index_of_interest)
            assert tree.hash() == proof.hash()



import hashlib
from scrypt import hash as scrypt_hash


def sha256d(b):
    return hashlib.sha256(hashlib.sha256(b).digest()).digest()


def scrypt(password, salt):
    # Why did we choose the scrypt factors that we did? Well... we at least bothered to read the abstract of the scrypt
    # paper, which is apparently more than the creators of Litecoin (and all its clones) did. As such, we tried to tune
    # the parameters using the super-scientific method of timing it our our local development machines, aiming for
    # 100ms.  N= 1 << 15 turned out to be the magic number, as guessed correctly by mr. Percival himself:
    # https://github.com/golang/go/issues/22082#issuecomment-332983728

    # Having said that, remember that ancient truth of crypto-currency: everything you see is shoddily built, since you
    # can always paper over your mistakes with technobabble. Present coin included.

    # More (fun) reading:
    # https://bitcoin.stackexchange.com/questions/36642/why-did-litecoin-choose-the-scrypt-factors-that-they-did

    # buflen 32 was chosen... because scrypt's output is going to through sha256d anyway, so no sense in a greater
    # output space
    return scrypt_hash(password, salt, N=1 << 15, r=8, p=1, buflen=32)


def blake2(b):
    return hashlib.blake2b(b, digest_size=32).digest()



import os
import ecdsa
import json

from .humans import human, computer
from .signing import SECP256k1PublicKey, SECP256k1Signature
from .datatypes import Input, Transaction, Output
from .coinstate import PKBalance


class AddressParseError(Exception):
    pass


class Wallet:

    def __init__(self, keypairs, unused_public_keys, public_key_annotations):
        self.keypairs = keypairs  # public => private

        self.unused_public_keys = unused_public_keys
        self.public_key_annotations = public_key_annotations

        self.spent_transaction_outputs = set()  # TODO save to disk too at some point.

    def __repr__(self):
        return "Wallet w/ %s keypairs" % len(self.keypairs)

    @classmethod
    def empty(cls):
        return cls({}, [], {})

    def __getitem__(self, public_key):
        return self.keypairs[public_key]

    def __contains__(self, public_key):
        return public_key in self.keypairs

    def get_annotated_public_key(self, annotation):
        public_key = self.unused_public_keys.pop()
        self.public_key_annotations[public_key] = annotation
        return public_key

    def generate_key(self):
        sk = ecdsa.SigningKey.generate(curve=ecdsa.SECP256k1)
        private_key = sk.to_string()  # deceptive naming: to_string() actually returns a bytes object
        public_key = sk.verifying_key.to_string()
        self.keypairs[public_key] = private_key
        self.unused_public_keys.append(public_key)

    def generate_keys(self, n=100):
        for i in range(n):
            self.generate_key()

    def dump(self, f):
        # The Simplest Thing That Could Possibly Work (though not the most secure)
        json.dump({
            "keypairs": {human(k): human(v) for (k, v) in self.keypairs.items()},
            "unused_public_keys": [human(e) for e in self.unused_public_keys],
            "public_key_annotations": {human(k): annotation for (k, annotation) in self.public_key_annotations.items()},
        }, f, indent=4)

    @classmethod
    def load(cls, f):
        d = json.load(f)

        return cls(
            keypairs={computer(k): computer(v) for (k, v) in d["keypairs"].items()},
            unused_public_keys=[computer(e) for e in d["unused_public_keys"]],
            public_key_annotations={computer(k): annotation for (k, annotation) in d["public_key_annotations"].items()},
        )

    def get_balance(self, coinstate):
        return sum(
            coinstate.public_key_balances_by_hash[coinstate.current_chain_hash].get(
                SECP256k1PublicKey(pk), PKBalance(0, [])).value
            for pk in self.keypairs.keys()
        )


def sign_transaction(wallet, unspent_transaction_outs, transaction):
    message = transaction.signable_equivalent().serialize()

    signed_inputs = []
    for input in transaction.signable_equivalent().inputs:
        if input.output_reference not in unspent_transaction_outs:
            raise Exception("Attempting to sign invalid transaction")

        output = unspent_transaction_outs[input.output_reference]

        if not isinstance(output.public_key, SECP256k1PublicKey):
            raise Exception("No idea how to sign this thing")

        if output.public_key.public_key not in wallet:
            raise Exception("Can't sign this; no known private key in wallet")

        private_key = wallet[output.public_key.public_key]

        sk = ecdsa.SigningKey.from_string(private_key, curve=ecdsa.SECP256k1)

        signed_inputs.append(Input(
            output_reference=input.output_reference,
            signature=SECP256k1Signature(sk.sign(message)),
        ))

    return Transaction(
        inputs=signed_inputs,
        outputs=transaction.outputs,
    )


def create_spend_transaction(wallet, coinstate, value, miners_fee, output_public_key, change_address):
    collected_value = 0
    inputs = []

    unspent_transaction_outs = coinstate.at_head.unspent_transaction_outs

    for public_key in wallet.keypairs.keys():
        if SECP256k1PublicKey(public_key) not in coinstate.at_head.public_key_balances:
            continue

        for output_reference in coinstate.at_head.public_key_balances[SECP256k1PublicKey(public_key)].output_references:
            if output_reference in wallet.spent_transaction_outputs:
                # in spent_transaction_outputs we keep track of those outputs that we've spent using this wallet (and
                # presumably broadcast) but which haven't made it into the chain yet.
                continue

            wallet.spent_transaction_outputs.add(output_reference)

            inputs.append(Input(output_reference, None))

            collected_value += unspent_transaction_outs[output_reference].value

            if collected_value >= value + miners_fee:
                outputs = [Output(value, output_public_key)]

                if collected_value != value + miners_fee:
                    outputs.append(Output(
                        collected_value - (value + miners_fee),
                        change_address,
                    ))

                return sign_transaction(wallet, unspent_transaction_outs, Transaction(inputs, outputs))

    raise Exception("Insufficient balance")


def save_wallet(wallet):
    # This manual handling of files is sure to create wallet file corruption at one point or another... oh well,
    # reliving the bitcoin experience one mistake at a time.

    with open("wallet.json.new", 'w') as f:
        wallet.dump(f)

    os.replace("wallet.json.new", "wallet.json")


def is_valid_address(full_skepticoin_address):
    try:
        parse_address(full_skepticoin_address)
        return True
    except AddressParseError:
        return False


def parse_address(full_skepticoin_address):
    if full_skepticoin_address[:3] != "SKE":
        raise AddressParseError()

    if full_skepticoin_address[-3:] != "PTI":
        raise AddressParseError()

    if len(full_skepticoin_address) != 3 + 128 + 3:
        raise AddressParseError()

    if not all(c in '0123456789abcdef' for c in full_skepticoin_address[3:-3]):
        raise AddressParseError()

    return computer(full_skepticoin_address[3:-3])



from .humans import computer

# In cryptocurrency tradition, the genesis block contains a newspaper headline that serves 2 purposes:
#
# [a] riling up the troops, i.e. convincing everyone that they're all in some kind of political project together.
# [b] proving that the blockchain is no older than a certain date (how else could you have produced the headline?).
#
# The second purpose is somewhat suspicious though, since:
#
# [1] given crypto's own set of assumptions, how long you spend calculating hashes shouldn't matter, all that matters is
#     the total (on average) of hashes being calculated.
# [2] the main threat against any crypto is not that the original creator spent some time mining coins, but that someone
#     else comes along with a _newer_ coin that's just as good or better than the old one. No newspaper headline can
#     guard against that.
#
# Thus, Skepticoin's genesis block only does [a].

just_believe_in_me = (
    b"You buy a piece of paradise\n"
    b"You buy a piece of me\n"
    b"I'll get you everything you wanted\n"
    b"I'll get you everything you need\n"
    b"Don't need to believe in hereafter\n"
    b"Just believe in me"
)

genesis_block_data = computer(
    '00000000000000000000000000000000000000000000000000000000000000000000616c35621abdf928185b74d57985cea7ff2d66ef318c58'
    'ec7ea8dd01ed089028604e7f3101000000000000000000000000000000000000000000000000000000000000000000003aea13176dbcbf6210'
    '55bdb3d6a138be4b73229d8584cb380e1dd1bbe1cedd42820000000000000000000000000000000000000000000000000000000000000000e3'
    '8ee41a6b0f6584fe8b95bd8c8d7b4d6db961fa5c2a6fafe72ea1533dd2838b0100010000000000000000000000000000000000000000000000'
    '000000000000000000000000000100000000ab596f75206275792061207069656365206f662070617261646973650a596f7520627579206120'
    '7069656365206f66206d650a49276c6c2067657420796f752065766572797468696e6720796f752077616e7465640a49276c6c206765742079'
    '6f752065766572797468696e6720796f75206e6565640a446f6e2774206e65656420746f2062656c6965766520696e20686572656166746572'
    '0a4a7573742062656c6965766520696e206d6501000000003b9aca0002aac3faad6ddc26ec4674328741498fe74bdb0d8e49a22473a02370e5'
    '3d69b0079819d5ac3f0cd36f25578eb042ad2a7b59f84a0b5f622e41ac982f478e8cb259'
)



"""
Tools to display hashes to humans.

We deviate from the bitcoin tradition, ostensibly inspired by shell games, of arbitrarily reverting bytes before
hexlifying them. We use plain old hexlify instead.
"""
from binascii import hexlify, unhexlify


def human(b):
    return hexlify(b).decode('utf-8')


def computer(s):
    return unhexlify(s.encode('utf-8'))


__all__ = ["human", "computer"]



import struct

from .humans import human
from .serialization import (
    safe_read,
    Serializable,
    stream_deserialize_list,
    stream_deserialize_vlq,
    stream_serialize_list,
    stream_serialize_vlq,
)
from .signing import Signature, PublicKey, SignableEquivalent
from .hash import sha256d
from .params import CHAIN_SAMPLE_TOTAL_SIZE


class OutputReference(Serializable):
    """Refer an output by its transaction hash and index into its list of outputs."""

    def __init__(self, hash, index):
        if not len(hash) == 32:
            raise ValueError('OutputReference hash must be 32 bytes.')

        if not (0 <= index <= 0xffffffff):
            raise ValueError('OutputReference index out of range.' % index)

        self.hash = hash
        self.index = index

    def __repr__(self):
        return "OutputReference(%s, %s)" % (human(self.hash), self.index)

    def __eq__(self, other):
        return self.hash == other.hash and self.index == other.index

    def __hash__(self):
        return hash((self.hash, self.index))

    @classmethod
    def stream_deserialize(cls, f):
        hash = safe_read(f, 32)
        (index,) = struct.unpack(b">I", safe_read(f, 4))
        return cls(hash, index)

    def stream_serialize(self, f):
        f.write(self.hash)
        f.write(struct.pack(b">I", self.index))

    def references_thin_air(self):
        return (self.hash == b'\x00' * 32) and (self.index == 0)


class Input(Serializable):
    """Input of a transaction"""

    def __init__(self, output_reference, signature):
        # bitcoin's sequence number (unused) is not reproduced here. See
        # https://bitcoin.stackexchange.com/questions/2025/what-is-txins-sequence

        self.output_reference = output_reference
        self.signature = signature

    def __repr__(self):
        return "Input(%s, %s)" % (self.output_reference, self.signature)

    def __eq__(self, other):
        return self.output_reference == other.output_reference and self.signature == other.signature

    @classmethod
    def stream_deserialize(cls, f):
        output_reference = OutputReference.stream_deserialize(f)
        signature = Signature.stream_deserialize(f)
        return cls(output_reference, signature)

    def stream_serialize(self, f):
        self.output_reference.stream_serialize(f)
        self.signature.stream_serialize(f)

    def signable_equivalent(self):
        # we simply replace the signature with a SignableEquivalent (which serializes to a type-byte)
        return Input(
            output_reference=self.output_reference,
            signature=SignableEquivalent()
        )


class Output(Serializable):
    """Output of a transaction."""

    def __init__(self, value, public_key):
        self.value = value
        self.public_key = public_key

    def __repr__(self):
        return "Output(%s, %s)" % (self.value, self.public_key)

    def __eq__(self, other):
        return self.value == other.value and self.public_key == other.public_key

    @classmethod
    def stream_deserialize(cls, f):
        (value,) = struct.unpack(b">Q", safe_read(f, 8))
        public_key = PublicKey.stream_deserialize(f)
        return cls(value, public_key)

    def stream_serialize(self, f):
        f.write(struct.pack(b">Q", self.value))
        self.public_key.stream_serialize(f)


class Transaction(Serializable):

    def __init__(self, inputs, outputs):
        self.version = 0  # reserved for future use; the class does not take this as a param.
        self.inputs = inputs
        self.outputs = outputs

    def __repr__(self):
        return "Transaction #%s" % human(self.hash())

    def __eq__(self, other):
        return self.inputs == other.inputs and self.outputs == other.outputs

    @classmethod
    def stream_deserialize(cls, f):
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version supports only version 0 transactions")

        inputs = stream_deserialize_list(f, Input)
        outputs = stream_deserialize_list(f, Output)
        return cls(inputs, outputs)

    def stream_serialize(self, f):
        f.write(struct.pack(b"B", self.version))
        stream_serialize_list(f, self.inputs)
        stream_serialize_list(f, self.outputs)

    def hash(self):
        return sha256d(self.serialize())

    def __hash__(self):
        return hash(self.hash())

    def signable_equivalent(self):
        # Because transactions contain (in their inputs) signatures, they cannot be signed as-is (you would need to know
        # the signature). So we sign a thing with the signatures taken out instead.
        return Transaction(
            inputs=[input.signable_equivalent() for input in self.inputs],
            outputs=self.outputs,
        )


class PowEvidence(Serializable):

    def __init__(self, summary_hash, chain_sample, block_hash):
        self.summary_hash = summary_hash
        self.chain_sample = chain_sample
        self.block_hash = block_hash

    def __repr__(self):
        return "PowEvidence(%s %s %s)" % (
            human(self.summary_hash),
            human(self.chain_sample),
            human(self.block_hash),
            )

    def __eq__(self, other):
        return (
            isinstance(other, PowEvidence) and
            self.summary_hash == other.summary_hash and
            self.chain_sample == other.chain_sample and
            self.block_hash == other.block_hash
        )

    @classmethod
    def stream_deserialize(cls, f):
        summary_hash = safe_read(f, 32)
        chain_sample = safe_read(f, CHAIN_SAMPLE_TOTAL_SIZE)
        block_hash = safe_read(f, 32)

        return cls(summary_hash, chain_sample, block_hash)

    def stream_serialize(self, f):
        f.write(self.summary_hash)
        f.write(self.chain_sample)
        f.write(self.block_hash)


class BlockSummary(Serializable):
    # akin to Bitcoin's BlockHeader. Our BlockHeader contains an PowEvidence also though, so we need an extra layer

    def __init__(self, height, previous_block_hash, merkle_root_hash, timestamp, target, nonce):
        # block height is included here to allow for partial validation of the POW in absence of the full block. (block
        # height is needed to calculate the blocks to sample from in chain_sample).
        self.height = height

        self.previous_block_hash = previous_block_hash
        self.merkle_root_hash = merkle_root_hash
        self.timestamp = timestamp

        # the `target`, i.e. the upper bound for the block's hash. We store this directly as a 32-byte rather than using
        # e.g. bitcoin's "bits" (which is yet another thing to understand/explain).
        self.target = target
        self.nonce = nonce  # it could be argued that this should be in PowEvidence

    def __repr__(self):
        return "BlockSummary #%s" % human(self.hash())

    def __eq__(self, other):
        return (
            isinstance(other, BlockSummary) and
            self.previous_block_hash == other.previous_block_hash and
            self.merkle_root_hash == other.merkle_root_hash and
            self.timestamp == other.timestamp and
            self.target == other.target and
            self.nonce == other.nonce
        )

    @classmethod
    def stream_deserialize(cls, f):
        height = stream_deserialize_vlq(f)
        previous_block_hash = safe_read(f, 32)
        merkle_root_hash = safe_read(f, 32)
        (timestamp,) = struct.unpack(b">I", safe_read(f, 4))
        target = safe_read(f, 32)
        (nonce,) = struct.unpack(b">I", safe_read(f, 4))

        return cls(height, previous_block_hash, merkle_root_hash, timestamp, target, nonce)

    def stream_serialize(self, f):
        stream_serialize_vlq(f, self.height)
        f.write(self.previous_block_hash)
        f.write(self.merkle_root_hash)
        f.write(struct.pack(b">I", self.timestamp))
        f.write(self.target)
        f.write(struct.pack(b">I", self.nonce))

    def hash(self):
        return sha256d(self.serialize())


class BlockHeader(Serializable):

    def __init__(self, summary, pow_evidence):
        self.version = 0
        self.summary = summary
        self.pow_evidence = pow_evidence

    def __repr__(self):
        return "BlockHeader #%s" % human(self.hash())

    def __eq__(self, other):
        return (
            isinstance(other, BlockHeader) and
            self.summary == other.summary and
            self.pow_evidence == other.pow_evidence
        )

    @classmethod
    def stream_deserialize(cls, f):
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version only supports version 0 blocks")

        summary = BlockSummary.stream_deserialize(f)
        pow_evidence = PowEvidence.stream_deserialize(f)

        return cls(summary, pow_evidence)

    def stream_serialize(self, f):
        f.write(struct.pack(b"B", self.version))

        self.summary.stream_serialize(f)
        self.pow_evidence.stream_serialize(f)

    def hash(self):
        return sha256d(self.serialize())


class Block(Serializable):
    def __init__(self, header, transactions):
        self.header = header
        self.transactions = transactions

    def __getattr__(self, attr):
        """convenience: merge header and summary's attributes into the Block's accessors"""
        if attr in ['version', 'summary', 'pow_evidence', 'hash']:
            return getattr(self.header, attr)

        if attr in ['height', 'previous_block_hash', 'merkle_root_hash', 'timestamp', 'target', 'nonce']:
            return getattr(self.header.summary, attr)

        raise AttributeError("'Block' object has no attribute '%s'" % attr)

    def __repr__(self):
        return "Block #%s" % human(self.header.hash())

    def __eq__(self, other):
        return (
            isinstance(other, Block) and
            self.header == other.header and
            # for valid blocks comparing transactions is superfluous but we don't make that assumption here
            self.transactions == other.transactions
        )

    @classmethod
    def stream_deserialize(cls, f):
        header = BlockHeader.stream_deserialize(f)
        transactions = stream_deserialize_list(f, Transaction)
        return cls(header, transactions)

    def stream_serialize(self, f):
        self.header.stream_serialize(f)
        stream_serialize_list(f, self.transactions)

    def get_total_work(self):
        return self.height  # TODO this is totally a placeholder :-D


__all__ = [
    'OutputReference',
    'Input',
    'Output',
    'Transaction',
    'PowEvidence',
    'BlockSummary',
    'BlockHeader',
    'Block',
]



from io import BytesIO
from skepticoin.serialization import stream_deserialize_vlq, stream_serialize_vlq


def test_vlq():
    test_cases = [0, 1, 42, 0x7f, 0x80, 0x2000, 0x3fff, 0x4000, 1234567890]
    for i in test_cases:
        f = BytesIO()
        stream_serialize_vlq(f, i)

        f.seek(0)
        assert stream_deserialize_vlq(f) == i



import immutables

from .humans import human, computer
from .params import (
    MAX_SASHIMI, MAX_BLOCK_SIZE, MAX_FUTURE_BLOCK_TIME, MAX_COINBASE_RANDOM_DATA_SIZE,
    BLOCKS_BETWEEN_TARGET_READJUSTMENT, DESIRED_TARGET_READJUSTMENT_TIMESPAN, SUBSIDY_HALVING_INTERVAL, INITIAL_SUBSIDY,
    CHAIN_SAMPLE_TOTAL_SIZE, CHAIN_SAMPLE_COUNT, CHAIN_SAMPLE_SIZE, INITIAL_TARGET
)
from .serialization import serialize_list
from .signing import CoinbaseData
from .merkletree import get_merkle_root
from .datatypes import OutputReference, Input, Output, Transaction, BlockSummary, PowEvidence, BlockHeader, Block
from .hash import scrypt, blake2
from .pow import select_n_k_length_slices_from_chain
from .coinstate import CoinState
from .cheating import KNOWN_HASHES, MAX_KNOWN_HASH_HEIGHT


# ## Section: Shared between Construction & Validation

def calc_merkle_root_hash(transactions):
    return get_merkle_root([transaction.hash() for transaction in transactions])


def calc_target(coinstate, height, current_timestamp, previous_block):
    if height % BLOCKS_BETWEEN_TARGET_READJUSTMENT == 0:
        interval_start_height = height - BLOCKS_BETWEEN_TARGET_READJUSTMENT
        interval_start_block = coinstate.block_by_height_by_hash[previous_block.hash()][interval_start_height]
        time_passed = current_timestamp - interval_start_block.timestamp
        return calculate_new_target(previous_block.target, time_passed)

    return previous_block.target


# ## Section: Construction of new Blocks

def construct_minable_summary_genesis(transactions, current_timestamp, nonce):
    return BlockSummary(
        height=0,
        previous_block_hash=b'\x00' * 32,
        merkle_root_hash=calc_merkle_root_hash(transactions),
        timestamp=current_timestamp,
        target=INITIAL_TARGET,
        nonce=nonce,
    )


def construct_minable_summary(coinstate, transactions, current_timestamp, nonce):
    if coinstate.current_chain_hash is None:
        return construct_minable_summary_genesis(transactions, current_timestamp, nonce)

    previous_block = coinstate.head()
    height = previous_block.height + 1
    return BlockSummary(
        height=height,
        previous_block_hash=coinstate.current_chain_hash,
        merkle_root_hash=calc_merkle_root_hash(transactions),
        timestamp=current_timestamp,
        target=calc_target(coinstate, height, current_timestamp, previous_block),
        nonce=nonce,
    )


def get_transaction_fee(transaction, unspent_transactions):
    """Money not explicitly claimed in a transaction will be available as a fee for the miner"""
    # should be called on valid transactions only (i.e. unspent output exists, no overspending)

    total_input_value = sum(
        unspent_transactions[input.output_reference].value
        for input in transaction.inputs
    )

    total_output_value = sum(output.value for output in transaction.outputs)

    return total_input_value - total_output_value


def get_block_fees(non_coinbase_transactions, unspent_transaction_outs):
    # Because intra-block cross-transaction-spending is illegal, no need to manipulate refetch unspent_transaction_outs
    return sum(get_transaction_fee(transaction, unspent_transaction_outs) for transaction in non_coinbase_transactions)


def get_block_subsidy(height):
    halvings = height // SUBSIDY_HALVING_INTERVAL

    if halvings >= 64:
        return 0

    return INITIAL_SUBSIDY // (2 ** halvings)


def construct_reference_to_thin_air():
    return OutputReference(b'\x00' * 32, 0)


def construct_coinbase_transaction(height, other_transactions, unspent_transaction_outs, signature, miner_public_key):
    subsidy = get_block_subsidy(height)
    fees = get_block_fees(other_transactions, unspent_transaction_outs)

    coinbase_data = CoinbaseData(
        height=height,
        signature=signature,
    )

    input = Input(
        output_reference=construct_reference_to_thin_air(),
        signature=coinbase_data,
    )

    output = Output(
        value=subsidy + fees,
        public_key=miner_public_key,
    )

    return Transaction(inputs=[input], outputs=[output])


def construct_pow_evidence(coinstate, summary, current_height, transactions):
    # Part 1 of the POW is to run scrypt. We put the most expensive operation first in an attempt to "up the ante"
    summary_hash = scrypt(summary.serialize(), current_height.to_bytes(8, byteorder='big'))

    # Part 2 of the POW is to prove that you have fast access to the whole blockchain. This is to help actual full nodes
    # vis-a-vis opportunistic miners.
    if current_height == 0:
        # in the genesis block we can't sample from the chain yet
        chain_sample = b'\00' * CHAIN_SAMPLE_TOTAL_SIZE
    else:
        get_block_by_height = lambda h: coinstate.block_by_height_by_hash[summary.previous_block_hash][h]  # noqa

        chain_sample = select_n_k_length_slices_from_chain(
            summary_hash, current_height, get_block_by_height, CHAIN_SAMPLE_COUNT, CHAIN_SAMPLE_SIZE)

    serialized_transactions = serialize_list(transactions)

    # Part 3 of the POW is to prove that the machine doing the work had access to the full list of transactions that
    # will be included in the block. The scenario this might guard against is that of malevolent parties renting mining
    # power to double-spend and/or mining pools gaining too much power and using it in ways that the members of the pool
    # would not agree to if they had seen the transactions.

    # This is a deviation from Bitcoin, in which the fixed length of the thing-to-mine is promoted as a feature, because
    # it does not create an incentive for miners to keep blocks short. To which we say: [1] why is it bad if there is at
    # least some incentive to keep blocks short, since this benefits the whole network? [2] we use a hashing algo that
    # is very fast in comparison to the main hashing algo (scrypt, above). We know this, because we spent 5 minutes
    # googling "fast cryptographic hash" and then picked the one which has native Python support. On 10MB blocks (many
    # times our actual block size) this hash runs in 0.02s on the developer's personal laptop (science!)
    block_hash = blake2(summary_hash + chain_sample + serialized_transactions)

    return PowEvidence(
        summary_hash=summary_hash,
        chain_sample=chain_sample,
        block_hash=block_hash,
    )


def construct_block_for_mining_genesis(
        non_coinbase_transactions, miner_public_key, current_timestamp, random_data, nonce):

    coinstate = CoinState.empty()
    current_height = 0

    unspent_transaction_outs = immutables.Map()

    coinbase_transaction = construct_coinbase_transaction(
        current_height, non_coinbase_transactions, unspent_transaction_outs, random_data, miner_public_key)

    transactions = [coinbase_transaction] + non_coinbase_transactions

    summary = construct_minable_summary(coinstate, transactions, current_timestamp, nonce)
    evidence = construct_pow_evidence(coinstate, summary, current_height, transactions)
    return Block(BlockHeader(summary, evidence), transactions)


def construct_block_for_mining(
        coinstate, non_coinbase_transactions, miner_public_key, current_timestamp, random_data, nonce):

    previous_block = coinstate.head()
    current_height = previous_block.height + 1

    unspent_transaction_outs = coinstate.unspent_transaction_outs_by_hash[coinstate.current_chain_hash]

    coinbase_transaction = construct_coinbase_transaction(
        current_height, non_coinbase_transactions, unspent_transaction_outs, random_data, miner_public_key)

    transactions = [coinbase_transaction] + non_coinbase_transactions

    summary = construct_minable_summary(coinstate, transactions, current_timestamp, nonce)
    evidence = construct_pow_evidence(coinstate, summary, current_height, transactions)
    return Block(BlockHeader(summary, evidence), transactions)

    # what would be nice to be able to verify independenly?
    # 1. calculation of summary_hash. (at cost of running scrypt once on the summary)
    # 2. calculation of any address byte individually (given the preceding bytes (if any) and a single block)
    # 3. calculation of bytes_from_block (given the current_block,


# ## Section: Validation


class ValidationError(Exception):
    pass


class ValidateTransactionError(ValidationError):
    pass


class ValidatePOWError(ValidationError):
    pass


class ValidateBlockHeaderError(ValidationError):
    pass


class ValidateBlockError(ValidationError):
    pass


def validate_sashimi_range(value):
    if not (0 < value <= MAX_SASHIMI):
        raise ValidationError("Value out of range.")


def validate_non_coinbase_transaction_by_itself(transaction):
    """Do those Transaction checks that can be done outside of the context a blockchain"""

    if len(transaction.inputs) == 0:
        raise ValidateTransactionError("No inputs")

    if len(transaction.outputs) == 0:
        raise ValidateTransactionError("No outputs")

    # block size is an upper bound for transaction size; there is no separate constant.
    if len(transaction.serialize()) > MAX_BLOCK_SIZE:
        raise ValidateTransactionError("transaction > MAX_BLOCK_SIZE")

    total_transaction_output_value = 0
    for output in transaction.outputs:
        validate_sashimi_range(output.value)
        total_transaction_output_value += output.value
    validate_sashimi_range(total_transaction_output_value)

    # Check for duplicate inputs; "in theory" this is redundant because each transaction output can be spent only once,
    # which is checked in validate_non_coinbase_transaction_in_coinstate. Some redundancy in checking might be useful
    # though, and checking it here allows us to catch such duplications even without the context of a full chain.
    output_references = set()
    for input in transaction.inputs:
        if input.output_reference in output_references:
            raise ValidateTransactionError("Single output_reference referenced more than once in single transaction.")
        output_references.add(input.output_reference)

    for input in transaction.inputs:
        if input.output_reference.references_thin_air():
            raise ValidateTransactionError("Coinbase-like null-reference in non-coinbase transaction.")

        if input.signature.is_not_signature():
            # as elsewhere: "in theory" this is redundant, because such a signature will never validate an output
            # anyway. checking it here allows us to catch such duplications even without the context of a full chain.
            raise ValidateTransactionError("Non-signature Signature class used where a real one is expected.")


def validate_signature_for_spend(input, previous_output, transaction):
    message = transaction.signable_equivalent().serialize()
    if not input.signature.validate(previous_output.public_key, message):
        raise ValidateTransactionError("Wrong signature for claimed output")


def validate_coinbase_transaction_by_itself(transaction):
    if not len(transaction.inputs) == 1:
        raise ValidateTransactionError("Coinbase transaction should have precisely 1 input")

    if not transaction.inputs[0].output_reference.references_thin_air():
        raise ValidateTransactionError("Coinbase must create its value out of thin air")

    if not isinstance(transaction.inputs[0].signature, CoinbaseData):
        raise ValidateTransactionError("A coinbase transaction should have CoinbaseData")

    if len(transaction.inputs[0].signature.signature) > MAX_COINBASE_RANDOM_DATA_SIZE:
        raise ValidateTransactionError("Random data > MAX_COINBASE_RANDOM_DATA_SIZE")


def validate_proof_of_work(hash, target):
    # N.B. just the final hash... without examining the evidence!

    if hash >= target:
        raise ValidatePOWError("hash >= target")


def validate_block_header_by_itself(block_header, current_timestamp):
    validate_proof_of_work(block_header.hash(), block_header.summary.target)

    if block_header.summary.timestamp > current_timestamp + MAX_FUTURE_BLOCK_TIME:
        raise ValidateBlockHeaderError("Block timestamp in the future")


def validate_no_duplicate_transactions(transactions):
    seen_transactions = set()
    for transaction in transactions:
        if transaction in seen_transactions:
            raise ValidateTransactionError("Duplicate transaction.")
        seen_transactions.add(transaction)


def validate_no_duplicate_output_references_in_transactions(transactions):
    seen_output_references = set()
    for transaction in transactions:
        for input in transaction.inputs:
            if input.output_reference in seen_output_references:
                raise ValidateTransactionError("Duplicate output_reference.")
            seen_output_references.add(input.output_reference)


def validate_block_by_itself(block, current_timestamp):
    validate_block_header_by_itself(block.header, current_timestamp)

    if len(block.transactions) == 0:
        raise ValidateBlockError("No transactions in block")

    if len(block.serialize()) > MAX_BLOCK_SIZE:
        raise ValidateBlockError("Block > MAX_BLOCK_SIZE")

    coinbase_transaction = block.transactions[0]

    validate_coinbase_transaction_by_itself(coinbase_transaction)

    if coinbase_transaction.inputs[0].signature.height != block.height:
        raise ValidateBlockError("block.height != coinbase.height")

    for transaction in block.transactions[1:]:
        validate_non_coinbase_transaction_by_itself(transaction)

    # Check for duplicate transactions; "in theory" this is redundant because each transaction output can be spent only
    # once,  which is checked in validate_non_coinbase_transaction_in_coinstate. Additionally, we don't suffer and from
    # Bitcoin's faulty (auto-duplicating) merkle-tree vulnerability. However, better safe than checking it here allows
    # us to catch such duplications even without the context of a full chain.
    validate_no_duplicate_transactions(block.transactions[1:])

    # We need to do this here because we the spending check based on coinstate only looks at the initial coinstate of
    # the block (i.e. the spending inside blocks isn't reflected in that check)
    validate_no_duplicate_output_references_in_transactions(block.transactions[1:])

    if block.header.summary.merkle_root_hash != calc_merkle_root_hash(block.transactions):
        raise ValidateBlockError("Incorrect merkle_root_hash")


def validate_coinbase_transaction_in_coinstate(transaction, block, coinstate):
    previous_block = coinstate.block_by_hash[block.header.summary.previous_block_hash]
    previous_height = previous_block.height
    calculated_current_height = previous_height + 1

    if block.height != calculated_current_height:
        raise ValidateBlockHeaderError("Block's reported height incorrect.")

    unspent_transaction_outs = coinstate.unspent_transaction_outs_by_hash[block.header.summary.previous_block_hash]
    fees = get_block_fees(block.transactions[1:], unspent_transaction_outs)
    subsidy = get_block_subsidy(block.height)

    if sum(output.value for output in transaction.outputs) > fees + subsidy:
        raise ValidateTransactionError('Transaction overspending (Coinbase)')


def validate_non_coinbase_transaction_in_coinstate(transaction, at_hash, coinstate):
    # Note that unspent_transaction_outs is fetched only once here, reflecting the state at the beginning of the block;
    # the implication is that spending money from another transaction in the same block is illegal. Though I'm sure
    # there are theoretical advantages in allowing it, the extra complexity isn't worth it. This also means we have to
    # check elsewhere (validate_block_by_itself) that a singe block cannot contain 2 transactions with the same
    # transaction output.
    unspent_transaction_outs = coinstate.unspent_transaction_outs_by_hash[at_hash]

    total_input_value = 0

    for input in transaction.inputs:
        if input.output_reference not in unspent_transaction_outs:
            raise ValidateTransactionError("input's output_reference does not exist as an unspent out")

        previous_output = unspent_transaction_outs[input.output_reference]

        # bitcoin has the concept of COINBASE_MATURITY here; we don't reproduce that idea here, shifting the
        # responsibility to the clients. Reasoning: is spending of newly minted coins really that different from
        # spending newly acquired coins?

        validate_signature_for_spend(input, previous_output, transaction)

        total_input_value += previous_output.value

    if sum(output.value for output in transaction.outputs) > total_input_value:
        raise ValidateTransactionError('Transaction overspending')


def calculate_new_target(previous_target, actual_time_passed):
    i_previous_target = int.from_bytes(previous_target, byteorder='big', signed=False)

    # multiplications first to avoid loss of precision; integer arithmetic only to avoid float-weirdness.
    result = (i_previous_target * actual_time_passed) // DESIRED_TARGET_READJUSTMENT_TIMESPAN

    if result > pow(2, 32 * 8) - 1:
        result = pow(2, 32 * 8) - 1  # TBH we have bigger problems if the target has become "anything goes", but still..

    return result.to_bytes(32, byteorder='big', signed=False)


def validate_block_summary_in_coinstate(block_summary, coinstate):
    if block_summary.previous_block_hash not in coinstate.block_by_hash:
        raise ValidateBlockHeaderError("previous_block_hash unknown: %s" % human(block_summary.previous_block_hash))

    previous_block = coinstate.block_by_hash[block_summary.previous_block_hash]

    if block_summary.timestamp <= previous_block.timestamp:
        # A simplification w.r.t. bitcoin, which accepts so little about reality that it presumes time cannot be
        # synchronized with greater precision than 2 full hours.
        raise ValidateBlockHeaderError("Timestamps must be strictly increasing.")

    previous_height = previous_block.height
    calculated_current_height = previous_height + 1

    calculated_target = calc_target(coinstate, calculated_current_height, block_summary.timestamp, previous_block)

    if block_summary.target != calculated_target:
        raise ValidateBlockHeaderError("Block's reported target incorrect")


def validate_block_in_coinstate(block, coinstate):
    if block.height <= MAX_KNOWN_HASH_HEIGHT:
        if block.height in KNOWN_HASHES:
            if block.hash() != computer(KNOWN_HASHES[block.height]):
                raise ValidationError("No forks allowed before block %s" % MAX_KNOWN_HASH_HEIGHT)

        # all in-coinstate validation is skipped for such blocks; this may lead to invalid blocks being accepted in your
        # local coinstate, but never beyond one of the checkpoints from KNOWN_HASHES
        return

    validate_block_summary_in_coinstate(block.header.summary, coinstate)

    reconstructed_evidence = construct_pow_evidence(coinstate, block.header.summary, block.height, block.transactions)
    if block.header.pow_evidence != reconstructed_evidence:
        raise ValidateBlockError("POW Evidence incorrect")

    coinbase_transaction = block.transactions[0]
    validate_coinbase_transaction_in_coinstate(coinbase_transaction, block, coinstate)

    for transaction in block.transactions[1:]:
        validate_non_coinbase_transaction_in_coinstate(transaction, block.previous_block_hash, coinstate)



from pathlib import Path
from decimal import Decimal
import random

from skepticoin.params import SASHIMI_PER_COIN
from skepticoin.consensus import construct_block_for_mining
from skepticoin.signing import SECP256k1PublicKey
from skepticoin.wallet import save_wallet
from skepticoin.utils import block_filename
from time import time

from .utils import (
    initialize_peers_file,
    create_chain_dir,
    read_chain_from_disk,
    open_or_init_wallet,
    start_networking_peer_in_background,
    check_for_fresh_chain,
    configure_logging_by_argv,
)


def main():
    configure_logging_by_argv()

    create_chain_dir()
    coinstate = read_chain_from_disk()
    wallet = open_or_init_wallet()
    initialize_peers_file()
    thread = start_networking_peer_in_background(coinstate)
    thread.local_peer.show_stats()

    if check_for_fresh_chain(thread):
        thread.local_peer.show_stats()

    print("Starting mining")

    try:
        print("Starting main loop")

        while True:
            public_key = wallet.get_annotated_public_key("reserved for potentially mined block")
            save_wallet(wallet)

            nonce = random.randrange(1 << 32)
            while True:
                coinstate, transactions = thread.local_peer.chain_manager.get_state()
                increasing_time = max(int(time()), coinstate.head().timestamp + 1)
                block = construct_block_for_mining(
                    coinstate, transactions, SECP256k1PublicKey(public_key), increasing_time, b'', nonce)
                nonce = (nonce + 1) % (1 << 32)
                if block.hash() < block.target:
                    break

            coinstate = coinstate.add_block(block, int(time()))
            with open(Path('chain') / block_filename(block), 'wb') as f:
                f.write(block.serialize())
            print("FOUND", block_filename(block))
            print("Wallet balance: %s skepticoin" % (wallet.get_balance(coinstate) / Decimal(SASHIMI_PER_COIN)))

            thread.local_peer.chain_manager.set_coinstate(coinstate)
            thread.local_peer.network_manager.broadcast_block(block)

    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    finally:
        print("Stopping networking thread")
        thread.stop()
        print("Waiting for networking thread to stop")
        thread.join()



import sys
import urllib.request
from pathlib import Path
import os
import tempfile
import logging

from skepticoin.datatypes import Block
from skepticoin.coinstate import CoinState
from skepticoin.networking.threading import NetworkingThread
from skepticoin.wallet import Wallet, save_wallet
from skepticoin.networking.utils import load_peers
from skepticoin.params import DESIRED_BLOCK_TIMESPAN

from time import time, sleep


def initialize_peers_file():
    if not os.path.isfile("peers.json"):
        print("Creating new peers.json")
        with urllib.request.urlopen("https://pastebin.com/raw/CcfPX9mS") as response:
            with open("peers.json", "wb") as f:
                f.write(response.read())


def create_chain_dir():
    if not os.path.exists('chain'):
        print("Created new directory for chain")
        os.makedirs('chain')


def check_for_fresh_chain(thread):
    # wait until your chain is no more than 20 typical block-sizes old before you start mining yourself
    waited = False
    try:
        while thread.local_peer.chain_manager.coinstate.head().timestamp + (20 * DESIRED_BLOCK_TIMESPAN) < time():
            waited = True
            thread.local_peer.show_stats()
            print("Waiting for fresh chain")
            sleep(10)

        while len(thread.local_peer.network_manager.get_active_peers()) == 0:
            waited = True
            thread.local_peer.show_stats()
            print("Waiting for peers")
            sleep(3)

    except KeyboardInterrupt:
        print("WAITING ABORTED... CONTINUING DESPITE FRESHNESS/CONNECTEDNESS!")

    return waited


def read_chain_from_disk():
    print("Reading chain from disk")
    coinstate = CoinState.zero()
    for filename in sorted(os.listdir('chain')):
        height = int(filename.split("-")[0])
        if height % 1000 == 0:
            print(filename)

        block = Block.stream_deserialize(open(Path('chain') / filename, 'rb'))
        coinstate = coinstate.add_block_no_validation(block)

    return coinstate


def open_or_init_wallet():
    if os.path.isfile("wallet.json"):
        wallet = Wallet.load(open("wallet.json", "r"))
    else:
        wallet = Wallet.empty()
        wallet.generate_keys(10_000)
        save_wallet(wallet)
        print("Created new wallet w/ 10.000 keys")

    return wallet


def start_networking_peer_in_background(coinstate):
    print("Starting networking peer in background")
    thread = NetworkingThread(coinstate, 2412)
    thread.local_peer.network_manager.disconnected_peers = load_peers()
    thread.start()
    return thread


def configure_logging_for_file():
    log_filename = Path(tempfile.gettempdir()) / ("skepticoin-networking-%s.log" % int(time()))
    FORMAT = '%(asctime)s %(message)s'
    logging.basicConfig(format=FORMAT, stream=open(log_filename, "w"), level=logging.INFO)


def configure_logging_for_stdout():
    FORMAT = '%(asctime)s %(message)s'
    logging.basicConfig(format=FORMAT, stream=sys.stdout, level=logging.INFO)


def configure_logging_by_argv():
    if "--log-networking-to-file" in sys.argv:
        configure_logging_for_file()

    if "--log-networking-to-stdout" in sys.argv:
        configure_logging_for_stdout()



"""
Proof of Waiste... I mean Work

In general simplicity is one of the main goals in the Skepticoin codebase. The less smoke and mirrors, the more closely
we can observe the trick. Here we deviate from that approach somewhat, for the following reasons:

* We don't want 0.0000001% of the bitcoin/litecoin mining power to be enough to do a successful 51% attack on our coin.
  If we actually manage to make someone angry, better make them put in some work before they can attack us.
* The parameters of the POW algorithm were "carefully chosen" to favor regular Joes on their laptops over evil miners on
  ASICs. However, see https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=8516911 for thoughts on how long that may
  last.
"""

from skepticoin.hash import sha256d


def select_block_height(input_hash, current_height):
    # we interpret the first 8 bytes of the input_hash as a number, and then modulo height. This should be enough for
    # the near future because (1 << 64) // (60 * 60 * 24 * 365) == 584_942_417_355 years if we use 1s blocks

    # current_height is the height of the block which hash we're mining/verifying

    base = int.from_bytes(input_hash[:8], byteorder='big', signed=False)
    return base % current_height


def select_block_slice(hash, serialized_block, length):
    # we interpret the next 4 bytes of the hash as a number, and then modulo block length. Support for up to 4GiB blocks

    base = int.from_bytes(hash[8:12], byteorder='big', signed=False)
    start = base % len(serialized_block)

    result = b""
    while len(result) < length:
        result += serialized_block[start:start + length - len(result)]
        start = 0

    return result


def select_slice_from_chain(input_hash, current_height, get_block_by_height, length):
    selected_block_height = select_block_height(input_hash, current_height)

    selected_block = get_block_by_height(selected_block_height)

    return select_block_slice(input_hash, selected_block.serialize(), length)


def select_n_k_length_slices_from_chain(starting_hash, current_height, get_block_by_height, n, k):
    result = []

    current_hash = starting_hash
    for i in range(n):
        b = select_slice_from_chain(current_hash, current_height, get_block_by_height, k)
        result.append(b)

        if i != n - 1:
            # presumed to be cheap. We also pick something different from scrypt, to be "multi hash"
            current_hash = sha256d(current_hash + b)

    return b"".join(result)



import argparse

from ..humans import human

from .utils import open_or_init_wallet, save_wallet


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("annotation", help="Some text to help you remember a meaning for this receive address.")
    args = parser.parse_args()

    wallet = open_or_init_wallet()
    public_key = wallet.get_annotated_public_key(args.annotation)
    save_wallet(wallet)

    print("SKE" + human(public_key) + "PTI")



import os
import argparse

from ptpython.repl import embed, run_config
from ptpython.entry_points.run_ptpython import get_config_and_history_file

from skepticoin.__version__ import __version__

import skepticoin.datatypes
import skepticoin.networking.messages
import skepticoin.signing
import skepticoin.humans
from skepticoin.params import SASHIMI_PER_COIN

from .utils import (
    initialize_peers_file,
    create_chain_dir,
    read_chain_from_disk,
    open_or_init_wallet,
    start_networking_peer_in_background,
)


class EverythingIsNone:
    def __getattr__(self, attr):
        return None


def main():
    config_file, history_file = get_config_and_history_file(EverythingIsNone())

    parser = argparse.ArgumentParser()
    parser.add_argument("--vi-mode", help="Vi mode", action="store_true")
    args = parser.parse_args()

    create_chain_dir()
    coinstate = read_chain_from_disk()
    wallet = open_or_init_wallet()
    initialize_peers_file()
    thread = start_networking_peer_in_background(coinstate)
    thread.local_peer.show_stats()

    print("Starting REPL, exit with exit()")
    try:
        locals = {
            'thread': thread,
            'local_peer': thread.local_peer,
            'show_stats': thread.local_peer.show_stats,
            'wallet': wallet,
            'get_coinstate': lambda: thread.local_peer.chain_manager.coinstate,
            'SASHIMI_PER_COIN': SASHIMI_PER_COIN,
        }

        for module in [skepticoin.datatypes, skepticoin.networking.messages, skepticoin.signing, skepticoin.humans]:
            for attr in module.__all__:
                locals[attr] = getattr(module, attr)

        def configure(repl) -> None:
            if os.path.exists(config_file):
                run_config(repl, config_file)
            else:
                # from https://github.com/prompt-toolkit/ptpython/blob/master/examples/ptpython_config/config.py
                # Ask for confirmation on exit.
                repl.confirm_exit = False

            # embedded in other applications.
            repl.title = "Skepticoin %s " % __version__

        embed(
            vi_mode=args.vi_mode,
            locals=locals,
            configure=configure,
            history_filename=history_file,
            patch_stdout=True,
        )

    finally:
        print("Stopping networking thread")
        thread.stop()
        print("Waiting for networking thread to stop")
        thread.join()



import struct
from ipaddress import IPv6Address

from ..serialization import (
    DeserializationError,
    safe_read,
    Serializable,
    stream_deserialize_list,
    stream_deserialize_vlq,
    stream_serialize_list,
    stream_serialize_vlq,
)
from ..datatypes import Block, BlockHeader, Transaction


MSG_HELLO = b'\x00\x00'
MSG_GET_BLOCKS = b'\x00\x01'
MSG_INVENTORY = b'\x00\x02'
MSG_GET_DATA = b'\x00\x03'
MSG_DATA = b'\x00\x04'
MSG_GET_PEERS = b'\x00\x05'
MSG_PEERS = b'\x00\x06'

DATA_BLOCK = b'\x00\x00'
DATA_HEADER = b'\x00\x01'
DATA_TRANSACTION = b'\x00\x02'


DATATYPES = {
    DATA_BLOCK: Block,
    DATA_HEADER: BlockHeader,
    DATA_TRANSACTION: Transaction,
}


class MessageHeader(Serializable):
    def __init__(self, timestamp, id, in_response_to, context):
        self.version = 0
        self.timestamp = timestamp
        self.id = id
        self.in_response_to = in_response_to
        self.context = context

    @classmethod
    def stream_deserialize(cls, f):
        # version is ignored... we parse everything as version 0, since that's the only thing we ourselves speak.
        # once other versions get introduced we should start paying attention to what's in here, and parse as much as we
        # know about
        (version,) = struct.unpack(b"B", safe_read(f, 1))

        (timestamp,) = struct.unpack(b">I", safe_read(f, 4))
        (id,) = struct.unpack(b">I", safe_read(f, 4))
        (in_response_to,) = struct.unpack(b">I", safe_read(f, 4))
        (context,) = struct.unpack(b">Q", safe_read(f, 8))

        safe_read(f, 32)  # reserved space for later versions

        return cls(timestamp, id, in_response_to, context)

    def stream_serialize(self, f):
        f.write(struct.pack(b"B", self.version))

        f.write(struct.pack(b">I", self.timestamp))
        f.write(struct.pack(b">I", self.id))
        f.write(struct.pack(b">I", self.in_response_to))
        f.write(struct.pack(b">Q", self.context))

        f.write(b'\x00' * 32)  # reserved space for later versions

    def format(self):
        return "t%010d-i%010d-r%010d-c%020d" % (self.timestamp, self.id, self.in_response_to, self.context)


class Message(Serializable):

    @classmethod
    def stream_deserialize(cls, f):
        type_indicator = safe_read(f, 2)

        if type_indicator == MSG_HELLO:
            return HelloMessage.stream_deserialize(f)

        if type_indicator == MSG_GET_BLOCKS:
            return GetBlocksMessage.stream_deserialize(f)

        if type_indicator == MSG_INVENTORY:
            return InventoryMessage.stream_deserialize(f)

        if type_indicator == MSG_GET_DATA:
            return GetDataMessage.stream_deserialize(f)

        if type_indicator == MSG_DATA:
            return DataMessage.stream_deserialize(f)

        if type_indicator == MSG_GET_PEERS:
            return GetPeersMessage.stream_deserialize(f)

        if type_indicator == MSG_PEERS:
            return PeersMessage.stream_deserialize(f)

        raise DeserializationError("Non-supported message type")


class SupportedVersion(Serializable):

    def __init__(self, version):
        self.version = version

    @classmethod
    def stream_deserialize(cls, f):
        (version,) = struct.unpack(b"B", safe_read(f, 1))
        return cls(version)

    def stream_serialize(self, f):
        f.write(struct.pack(b"B", self.version))


class HelloMessage(Message):

    def __init__(self, supported_versions, your_ip_address, your_port, my_ip_address, my_port, nonce, user_agent):
        self.version = 0
        self.supported_versions = supported_versions

        # capabilities? (bitcoin calls this "services")... for now we imply that everyone is a full node.

        self.your_ip_address = your_ip_address
        self.your_port = your_port

        self.my_ip_address = my_ip_address
        self.my_port = my_port

        self.nonce = nonce

        self.user_agent = user_agent

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.

        # version is ignored... we parse everything as version 0, since that's the only thing we ourselves speak.
        # once other versions get introduced we should start paying attention to what's in here, and parse as much as we
        # know about.
        (version,) = struct.unpack(b"B", safe_read(f, 1))

        your_ip_address = IPv6Address(safe_read(f, 16))
        (your_port,) = struct.unpack(b">H", safe_read(f, 2))

        my_ip_address = IPv6Address(safe_read(f, 16))
        (my_port,) = struct.unpack(b">H", safe_read(f, 2))

        (nonce,) = struct.unpack(b">I", safe_read(f, 4))

        (ua_length,) = struct.unpack(b"B", safe_read(f, 1))
        user_agent = safe_read(f, ua_length)

        supported_versions = stream_deserialize_list(f, SupportedVersion)

        safe_read(f, 256)  # reserved space for later versions

        return cls(supported_versions, your_ip_address, your_port, my_ip_address, my_port, nonce, user_agent)

    def stream_serialize(self, f):
        f.write(MSG_HELLO)
        f.write(struct.pack(b"B", self.version))

        f.write(self.your_ip_address.packed)
        f.write(struct.pack(b">H", self.your_port))

        f.write(self.my_ip_address.packed)
        f.write(struct.pack(b">H", self.my_port))

        f.write(struct.pack(b">I", self.nonce))

        f.write(struct.pack(b"B", len(self.user_agent)))
        f.write(self.user_agent)

        stream_serialize_list(f, self.supported_versions)

        f.write(b'\x00' * 256)  # reserved space for later versions


class GetBlocksMessage(Message):

    def __init__(self, potential_start_hashes, stop_hash=b'\x00' * 32):
        self.version = 0

        self.potential_start_hashes = potential_start_hashes
        self.stop_hash = stop_hash

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version supports only version 0 GetBlocksMessage")

        potential_start_hashes = []
        length = stream_deserialize_vlq(f)
        for i in range(length):
            potential_start_hashes.append(safe_read(f, 32))

        stop_hash = safe_read(f, 32)
        return cls(potential_start_hashes, stop_hash)

    def stream_serialize(self, f):
        f.write(MSG_GET_BLOCKS)
        f.write(struct.pack(b"B", self.version))

        stream_serialize_vlq(f, len(self.potential_start_hashes))
        for h in self.potential_start_hashes:
            f.write(h)

        f.write(self.stop_hash)


class InventoryItem(Serializable):

    def __init__(self, data_type, hash):
        self.data_type = data_type
        self.hash = hash

    @classmethod
    def stream_deserialize(cls, f):
        data_type = safe_read(f, 2)
        hash = safe_read(f, 32)
        return cls(data_type, hash)

    def stream_serialize(self, f):
        f.write(self.data_type)
        f.write(self.hash)


class InventoryMessage(Message):

    def __init__(self, items):
        self.version = 0
        self.items = items

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version supports only version 0 InventoryMessage")

        items = stream_deserialize_list(f, InventoryItem)
        return cls(items)

    def stream_serialize(self, f):
        f.write(MSG_INVENTORY)
        f.write(struct.pack(b"B", self.version))

        stream_serialize_list(f, self.items)


class GetDataMessage(Message):

    def __init__(self, data_type, hash):
        self.version = 0
        self.data_type = data_type
        self.hash = hash

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version supports only version 0 GetDataMessage")

        data_type = safe_read(f, 2)
        hash = safe_read(f, 32)

        return cls(data_type, hash)

    def stream_serialize(self, f):
        f.write(MSG_GET_DATA)
        f.write(struct.pack(b"B", self.version))

        f.write(self.data_type)
        f.write(self.hash)


class DataMessage(Message):

    def __init__(self, data_type, data):
        self.version = 0

        self.data_type = data_type
        self.data = data

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version supports only version 0 DataMessage")

        data_type = safe_read(f, 2)

        clz = DATATYPES[data_type]
        data = clz.stream_deserialize(f)

        return cls(data_type, data)

    def stream_serialize(self, f):
        f.write(MSG_DATA)
        f.write(struct.pack(b"B", self.version))

        f.write(self.data_type)
        self.data.stream_serialize(f)


class GetPeersMessage(Message):

    def __init__(self):
        self.version = 0

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version supports only version 0 GetPeersMessage")

        return cls()

    def stream_serialize(self, f):
        f.write(MSG_GET_PEERS)
        f.write(struct.pack(b"B", self.version))


class Peer(Serializable):

    def __init__(self, last_seen_at, ip_address, port):
        self.last_seen_at = last_seen_at
        self.ip_address = ip_address
        self.port = port

    @classmethod
    def stream_deserialize(cls, f):
        (last_seen_at,) = struct.unpack(b">I", safe_read(f, 4))
        ip_address = IPv6Address(safe_read(f, 16))
        (port,) = struct.unpack(b">H", safe_read(f, 2))

        return cls(last_seen_at, ip_address, port)

    def stream_serialize(self, f):
        f.write(struct.pack(b">I", self.last_seen_at))
        f.write(self.ip_address.packed)
        f.write(struct.pack(b">H", self.port))


class PeersMessage(Message):

    def __init__(self, peers):
        self.version = 0
        self.peers = peers

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        if safe_read(f, 1) != b'\x00':
            raise ValueError("Current version supports only version 0 GetPeersMessage")

        peers = stream_deserialize_list(f, Peer)

        return cls(peers)

    def stream_serialize(self, f):
        f.write(MSG_PEERS)
        f.write(struct.pack(b"B", self.version))
        stream_serialize_list(f, self.peers)


__all__ = [
    "MessageHeader",
    "Message",
    "SupportedVersion",
    "HelloMessage",
    "GetBlocksMessage",
    "InventoryItem",
    "InventoryMessage",
    "GetDataMessage",
    "DataMessage",
    "GetPeersMessage",
    "Peer",
    "PeersMessage",
]



"""
Maximum removal of smoke and mirrors. This means:

* No fancy (though mostly unused) scripts here as in bitcoin. We stick to the core business: claiming by signing.
* There's no equivalent of a "bitcoin addresses", P2PKH or P2SH: just public keys and signatures.

One byte is reserved as a type_indicator if we ever want to offer more options (compressed public keys anyone?)
"""
import struct
import ecdsa  # NOTE "This library was not designed with security in mind."

from .humans import human
from .serialization import Serializable, DeserializationError, safe_read

TYPE_SIGNABLE_EQUIVALENT = b'\x00'
TYPE_COINBASE_DATA = b'\x01'
TYPE_SECP256k1 = b'\x02'


class PublicKey(Serializable):

    @classmethod
    def stream_deserialize(cls, f):
        type_indicator = safe_read(f, 1)

        if type_indicator == TYPE_SECP256k1:
            return SECP256k1PublicKey.stream_deserialize(f)

        raise DeserializationError("Non-supported public key type.")


class SECP256k1PublicKey(PublicKey):
    """We use the same curve as bitcoin because why not. Remember: the NIST curves were chosen by the lizard people!"""

    def __init__(self, public_key):
        if not len(public_key) == 64:
            raise ValueError('SECP256k1 public key must be 64 bytes.')

        self.public_key = public_key

    def __repr__(self):
        return "SECP256k1 Public Key %s" % human(self.public_key)

    def __eq__(self, other):
        return isinstance(other, SECP256k1PublicKey) and self.public_key == other.public_key

    def __hash__(self):
        return hash(self.serialize())

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        public_key = safe_read(f, 64)
        return cls(public_key)

    def stream_serialize(self, f):
        f.write(TYPE_SECP256k1)
        f.write(self.public_key)

    def validate(self, signature, message):
        if not isinstance(signature, SECP256k1Signature):
            return False

        vk = ecdsa.VerifyingKey.from_string(self.public_key, curve=ecdsa.SECP256k1)
        try:
            vk.verify(signature.signature, message)
            return True
        except ecdsa.keys.BadSignatureError:
            return False


class Signature(Serializable):

    @classmethod
    def stream_deserialize(cls, f):
        type_indicator = safe_read(f, 1)

        if type_indicator == TYPE_SIGNABLE_EQUIVALENT:
            return SignableEquivalent.stream_deserialize(f)

        if type_indicator == TYPE_COINBASE_DATA:
            return CoinbaseData.stream_deserialize(f)

        if type_indicator == TYPE_SECP256k1:
            return SECP256k1Signature.stream_deserialize(f)

        raise DeserializationError("Non-supported signature type.")

    def is_not_signature(self):
        """In various places where signatures are expected, special-meaning placeholders can occur instead. Signatures
        that may actually be used to verify public keys should return False here."""
        return True


class SignableEquivalent(Signature):
    """SignableEquivalent: when signing transactions you can't sign your own signature."""

    def __repr__(self):
        return 'SignableEquivalent()'

    def __eq__(self, other):
        return isinstance(other, SignableEquivalent)

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        return cls()

    def stream_serialize(self, f):
        f.write(TYPE_SIGNABLE_EQUIVALENT)


class CoinbaseData(Signature):
    """In Coinbase transactions, some random data takes the place of the signature. This may be used by miners to
    introduce extra randomness if the nonce is not enough, or to include pseudo-polical messages."""

    def __init__(self, height, signature):
        if not (0 <= height <= 0xffffffff):
            raise ValueError('CoinbaseData height out of range.' % height)

        if len(signature) > 256:
            raise ValueError("Unserializable CoinbaseData")

        # height is included to make guarantee Transaction uniqueness. (without height, mining a block with a single
        # coinbase transaction with the same output address twice would lead to non-uniqueness).
        self.height = height
        self.signature = signature

    def __repr__(self):
        if all([32 <= b < 127 for b in self.signature]):
            return 'CoinbaseData(%s, "%s")' % (self.height, str(self.signature, encoding="ascii"))
        return 'CoinbaseData(%s, #%s)' % (self.height, human(self.signature))

    def __eq__(self, other):
        return isinstance(other, CoinbaseData) and self.signature == other.signature

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        (height,) = struct.unpack(b">I", safe_read(f, 4))
        (length,) = struct.unpack(b"B", safe_read(f, 1))
        signature = safe_read(f, length)
        return cls(height, signature)

    def stream_serialize(self, f):
        f.write(TYPE_COINBASE_DATA)
        f.write(struct.pack(b">I", self.height))
        f.write(struct.pack(b"B", len(self.signature)))
        f.write(self.signature)


class SECP256k1Signature(Signature):

    def __init__(self, signature):
        if not len(signature) == 64:
            raise ValueError('SECP256k1 signature must be 64 bytes.')

        self.signature = signature

    def __repr__(self):
        return "SECP256k1 Signature %s" % human(self.signature)

    def __eq__(self, other):
        return isinstance(other, SECP256k1Signature) and self.signature == other.signature

    @classmethod
    def stream_deserialize(cls, f):
        # type_indicator has been read already by the superclass at this point.
        signature = safe_read(f, 64)
        return cls(signature)

    def stream_serialize(self, f):
        f.write(TYPE_SECP256k1)
        f.write(self.signature)

    def validate(self, public_key, message):
        if not isinstance(public_key, SECP256k1PublicKey):
            return False

        return public_key.validate(self, message)

    def is_not_signature(self):
        return False


__all__ = [
    "PublicKey",
    "SECP256k1PublicKey",
    "Signature",
    "SignableEquivalent",
    "CoinbaseData",
    "SECP256k1Signature",
]



from .humans import human


def block_filename(block):
    return "%08d-%s" % (block.height, human(block.hash()))


def calc_work(target):
    return pow(2, 32 * 8) // int.from_bytes(target, byteorder="big", signed=False)



from skepticoin.signing import SECP256k1Signature, SECP256k1PublicKey
from skepticoin.datatypes import (
    Block,
    BlockHeader,
    BlockSummary,
    Input,
    Output,
    OutputReference,
    PowEvidence,
    Transaction,
)
from skepticoin.humans import computer


def serialize_and_deserialize(thing):
    other_thing = type(thing).deserialize(thing.serialize())
    assert thing == other_thing


def test_output_reference_serialization():
    ref = OutputReference(b"a" * 32, 1234)
    serialize_and_deserialize(ref)


def test_input_serialization():
    inp = Input(
        output_reference=OutputReference(b"b" * 32, 1234),
        signature=SECP256k1Signature(b"b" * 64),
    )
    serialize_and_deserialize(inp)


def test_output_serialization():
    out = Output(
        value=1582,
        public_key=SECP256k1PublicKey(b"g" * 64),
    )
    serialize_and_deserialize(out)


def test_transaction_serialization():
    trans = Transaction(
        inputs=[Input(output_reference=OutputReference(b"b" * 32, 1234), signature=SECP256k1Signature(b"b" * 64))],
        outputs=[Output(value=1582, public_key=SECP256k1PublicKey(b"g" * 64))],
    )
    serialize_and_deserialize(trans)


def test_transaction_repr():
    trans = Transaction(
        inputs=[Input(output_reference=OutputReference(b"b" * 32, 1234), signature=SECP256k1Signature(b"b" * 64))],
        outputs=[Output(value=1582, public_key=SECP256k1PublicKey(b"g" * 64))],
    )

    assert repr(trans) == "Transaction #4025f3f13790dc96d857562dabcdd257ee9dfd95ce126e11d8cbbe64ac1bbec4"


example_block_summary = BlockSummary(
    height=445,
    previous_block_hash=b"a" * 32,
    merkle_root_hash=b"b" * 32,
    timestamp=1231006505,
    target=computer("00000000ffff0000000000000000000000000000000000000000000000000000"),
    nonce=1234,
)


def test_block_summary_serialization():
    serialize_and_deserialize(example_block_summary)


example_pow_evidence = PowEvidence(
    summary_hash=b"d" * 32,
    chain_sample=b"e" * 32,
    block_hash=b"f" * 32,
)


def test_pow_evidence_serialization():
    serialize_and_deserialize(example_pow_evidence)


def test_block_header_serialization():
    bh = BlockHeader(
        summary=example_block_summary,
        pow_evidence=example_pow_evidence,
    )
    serialize_and_deserialize(bh)


def test_block_serialization():
    block = Block(
        header=BlockHeader(
            summary=example_block_summary,
            pow_evidence=example_pow_evidence,
        ),
        transactions=[Transaction(
            inputs=[Input(output_reference=OutputReference(b"b" * 32, 1234), signature=SECP256k1Signature(b"b" * 64))],
            outputs=[Output(value=1582, public_key=SECP256k1PublicKey(b"g" * 64))],
        )] * 2,
    )

    serialize_and_deserialize(block)



import immutables
from collections import namedtuple

from .datatypes import OutputReference, Block
from .humans import human
from .genesis import genesis_block_data


PKBalance = namedtuple('PKBalance', ['value', 'output_references'])


def uto_apply_transaction(unspent_transaction_outs, transaction, is_coinbase):
    with unspent_transaction_outs.mutate() as mutable_unspent_transaction_outs:

        # for coinbase we must skip the input-removal because the input references "thin air" rather than an output.
        if not is_coinbase:
            for input in transaction.inputs:
                # we don't explicitly check that the transaction is spendable; inadvertant violation of that expectation
                # will lead to an application-crash (which is preferable over the alternative: double-spending).
                del mutable_unspent_transaction_outs[input.output_reference]

        for i, output in enumerate(transaction.outputs):
            output_reference = OutputReference(transaction.hash(), i)
            mutable_unspent_transaction_outs[output_reference] = output

        return mutable_unspent_transaction_outs.finish()


def uto_apply_block(unspent_transaction_outs, block):
    unspent_transaction_outs = uto_apply_transaction(unspent_transaction_outs, block.transactions[0], is_coinbase=True)
    for transaction in block.transactions[1:]:
        unspent_transaction_outs = uto_apply_transaction(unspent_transaction_outs, transaction, is_coinbase=False)
    return unspent_transaction_outs


def pkb_apply_transaction(unspent_transaction_outs, public_key_balances, transaction, is_coinbase):
    with public_key_balances.mutate() as mutable_public_key_balances:
        # for coinbase we must skip the input-removal because the input references "thin air" rather than an output.
        if not is_coinbase:
            for input in transaction.inputs:
                previously_unspent_output = unspent_transaction_outs[input.output_reference]

                public_key = previously_unspent_output.public_key
                mutable_public_key_balances[public_key] = PKBalance(
                    mutable_public_key_balances[public_key].value - previously_unspent_output.value,
                    [to for to in mutable_public_key_balances[public_key].output_references
                     if to != input.output_reference]
                )

        for i, output in enumerate(transaction.outputs):
            output_reference = OutputReference(transaction.hash(), i)

            if output.public_key not in mutable_public_key_balances:
                mutable_public_key_balances[output.public_key] = PKBalance(0, [])

            mutable_public_key_balances[output.public_key] = PKBalance(
                mutable_public_key_balances[output.public_key].value + output.value,
                mutable_public_key_balances[output.public_key].output_references + [output_reference],
            )

        return mutable_public_key_balances.finish()


def pkb_apply_block(unspent_transaction_outs, public_key_balances, block):
    # unspent_transaction_outs is used as a "reference" only (for looking up outputs); note that we never have to update
    # that reference inside this function, because intra-block spending is invalid per the consensus.

    public_key_balances = pkb_apply_transaction(
        unspent_transaction_outs, public_key_balances, block.transactions[0], is_coinbase=True)

    for transaction in block.transactions[1:]:
        public_key_balances = pkb_apply_transaction(unspent_transaction_outs, public_key_balances, transaction, False)

    return public_key_balances


class CoinState:
    def __init__(self, block_by_hash, unspent_transaction_outs_by_hash, block_by_height_by_hash, heads,
                 current_chain_hash, public_key_balances_by_hash):

        self.block_by_hash = block_by_hash

        # block_hash -> (OutputReference -> Output)
        self.unspent_transaction_outs_by_hash = unspent_transaction_outs_by_hash

        # given a hash to a current head, return a dictionary in which you can look up by height.
        self.block_by_height_by_hash = block_by_height_by_hash

        self.heads = heads  # hash=>block ... but restricted to blocks w/o children.
        self.current_chain_hash = current_chain_hash

        # block_hash -> (public_key -> (value, [OutputReference]))
        self.public_key_balances_by_hash = public_key_balances_by_hash

    def __repr__(self):
        if self.current_chain_hash is None:
            return "CoinState @ empty"
        return "CoinState @ %s (h. %s) w/ %s heads" % (
            human(self.current_chain_hash), self.head().height, len(self.heads))

    @classmethod
    def empty(cls):
        return cls(
            block_by_hash=immutables.Map(),
            unspent_transaction_outs_by_hash=immutables.Map(),
            block_by_height_by_hash=immutables.Map(),
            heads=immutables.Map(),
            current_chain_hash=None,
            public_key_balances_by_hash=immutables.Map(),
        )

    @classmethod
    def zero(cls):
        e = cls.empty()
        return e.add_block_no_validation(Block.deserialize(genesis_block_data))

    def add_block(self, block, current_timestamp):
        from skepticoin.consensus import validate_block_by_itself, validate_block_in_coinstate
        validate_block_by_itself(block, current_timestamp)
        validate_block_in_coinstate(block, self)

        return self.add_block_no_validation(block)

    def add_block_no_validation(self, block):
        if block.previous_block_hash == b'\00' * 32:
            unspent_transaction_outs = immutables.Map()
            public_key_balances = immutables.Map()
        else:
            unspent_transaction_outs = self.unspent_transaction_outs_by_hash[block.previous_block_hash]
            public_key_balances = self.public_key_balances_by_hash[block.previous_block_hash]

        public_key_balances = pkb_apply_block(unspent_transaction_outs, public_key_balances, block)
        # NOTE: ordering matters b/c we assign to unspent_transaction_outs; perhaps just distinguish?
        unspent_transaction_outs = uto_apply_block(unspent_transaction_outs, block)

        block_by_hash = self.block_by_hash.set(block.hash(), block)

        # TODO pruning of unspent_transaction_outs_by_hash on e.g. max delta-height; because it is used as part of
        # validation, such an approach implies a limit (to be implemented in validation) on how old a fork may be w.r.t.
        # the current height if it is to be considered at all.
        unspent_transaction_outs_by_hash = self.unspent_transaction_outs_by_hash.set(
            block.hash(), unspent_transaction_outs)

        public_key_balances_by_hash = self.public_key_balances_by_hash.set(
            block.hash(), public_key_balances)

        if block.previous_block_hash == b'\00' * 32:
            block_by_height_by_hash = immutables.Map({block.hash(): immutables.Map({0: block})})
        else:
            block_by_height = self.block_by_height_by_hash[block.previous_block_hash]
            block_by_height = block_by_height.set(block.height, block)
            block_by_height_by_hash = self.block_by_height_by_hash.set(block.hash(), block_by_height)

        with self.heads.mutate() as mutable_heads:
            if block.previous_block_hash in mutable_heads:
                del mutable_heads[block.header.summary.previous_block_hash]

            mutable_heads[block.hash()] = block

            heads = mutable_heads.finish()

        if self.current_chain_hash is None or self.current_chain_hash == block.previous_block_hash:
            # base case / simple moving ahead
            current_chain_hash = block.hash()

        # what should be the current_chain_hash in the case of forks?
        # we compare total work, using striclty greater: this means that when there is a fork in the chain, ties are
        # broken based on first-come-first serve. I'm sure someone wrote a paper on how this is optimal.
        elif block.get_total_work() > self.block_by_hash[self.current_chain_hash].get_total_work():
            current_chain_hash = block.hash()
        else:
            current_chain_hash = self.current_chain_hash  # a fork, but the most recently added block is non-current

        return CoinState(
            block_by_hash=block_by_hash,
            unspent_transaction_outs_by_hash=unspent_transaction_outs_by_hash,
            block_by_height_by_hash=block_by_height_by_hash,
            heads=heads,
            current_chain_hash=current_chain_hash,
            public_key_balances_by_hash=public_key_balances_by_hash,
        )

    def head(self):
        return self.block_by_hash[self.current_chain_hash]

    def by_height_at_head(self):
        # TODO just use 'at_head'
        return self.block_by_height_by_hash[self.current_chain_hash]

    @property
    def at_head(self):

        class AtHead:
            @property
            def unspent_transaction_outs(inner_self):
                return self.unspent_transaction_outs_by_hash[self.current_chain_hash]

            @property
            def block_by_height(inner_self):
                return self.block_by_height_by_hash[self.current_chain_hash]

            @property
            def public_key_balances(inner_self):
                return self.public_key_balances_by_hash[self.current_chain_hash]

        return AtHead()

    def forks(self):
        def _find_lca_with_main(block):
            # while block.hash() not in block_by_hash_by_hash  ... we have no such datastructure yet, so instead:

            while (block.height not in self.by_height_at_head() or
                    self.by_height_at_head()[block.height].hash() != block.hash()):

                block = self.block_by_hash[block.previous_block_hash]

            return block

        return [(head, _find_lca_with_main(head)) for head in self.heads.values()]



PORT = 2412
TIME_BETWEEN_CONNECTION_ATTEMPTS = 10

GET_PEERS_INTERVAL = 60 * 30

MAX_MESSAGE_SIZE = 32 * 1024 * 1024

MAX_IBD_PEERS = 1
IBD_PEER_TIMEOUT = 60

GET_BLOCKS_INVENTORY_SIZE = 500

SWITCH_TO_ACTIVE_MODE_TIMEOUT = 5 * 60  # if your chain is 5 minutes old, start querying for blocks actively
EMPTY_INVENTORY_BACKOFF = 60  # wait this long before asking a node about inventory again on an empty response



from setuptools import setup, find_packages

setup(
    name="skepticoin",
    description="The Coin for Non-Believers",
    long_description=open("README.md", 'r').read(),
    long_description_content_type='text/markdown',

    author="Sashimi Houdini",
    url="https://github.com/skepticoin/skepticoin/",

    install_requires=[
        "scrypt>=0.8.17",
        "ecdsa>=0.16.1",
        "immutables>=0.15",
        "ptpython",
    ],

    packages=find_packages(),

    setup_requires=["setuptools_scm"],
    use_scm_version={
        "write_to": "skepticoin/scmversion.py",
        "write_to_template": "__version__ = '{version}'\n",
    },

    include_package_data=True,

    entry_points={
        'console_scripts': [
            'skepticoin-version=skepticoin.scripts.version:main',
            'skepticoin-mine=skepticoin.scripts.mine:main',
            'skepticoin-receive=skepticoin.scripts.receive:main',
            'skepticoin-send=skepticoin.scripts.send:main',
            'skepticoin-repl=skepticoin.scripts.repl:main',
        ],
    },

    license="BSD-3-Clause",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',

        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: BSD License',

        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',

        'Programming Language :: Python',
    ],
)



from time import sleep
import argparse

from skepticoin.params import SASHIMI_PER_COIN
from skepticoin.signing import SECP256k1PublicKey
from skepticoin.wallet import save_wallet
from skepticoin.wallet import is_valid_address, parse_address, create_spend_transaction

from .utils import (
    initialize_peers_file,
    create_chain_dir,
    read_chain_from_disk,
    open_or_init_wallet,
    start_networking_peer_in_background,
    check_for_fresh_chain,
)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("amount", help="The amount of to send", type=int)
    parser.add_argument("denomination", help="'skepticoin' or 'sashimi'", choices=['skepticoin', 'sashimi'])
    parser.add_argument("address", help="The address to send to")
    args = parser.parse_args()

    value = args.amount * (SASHIMI_PER_COIN if args.denomination == 'skepticoin' else 1)

    if not is_valid_address(args.address):
        print("Invalid address")
        return

    create_chain_dir()
    coinstate = read_chain_from_disk()
    wallet = open_or_init_wallet()
    initialize_peers_file()
    thread = start_networking_peer_in_background(coinstate)

    try:
        # we need a fresh chain because our wallet doesn't track spending/receiving, so we need to look at the real
        # blockchain to know what we can spend.
        check_for_fresh_chain(thread)
        print("Chain up to date")

        target_address = SECP256k1PublicKey(parse_address(args.address))
        change_address = SECP256k1PublicKey(wallet.get_annotated_public_key("change"))
        save_wallet(wallet)

        transaction = create_spend_transaction(
            wallet,
            coinstate,
            value,
            0,  # we'll get to paying fees later
            target_address,
            change_address,
        )

        print("Broadcasting transaction on the network", transaction)
        thread.local_peer.network_manager.broadcast_transaction(transaction)

        print("Monitoring...")

        while True:
            sleep(5)

            # it's late and I'm too lazy for the efficient & correct implementation.
            coinstate = thread.local_peer.chain_manager.coinstate
            max_height = coinstate.head().height

            for i in range(10):
                block = coinstate.by_height_at_head()[max(max_height - i, 0)]
                if transaction in block.transactions:
                    print("Transaction confirmed at", block.height, "with", i, "confirmation blocks")

                    if i == 6:  # this is the magic number of confirmations according to the "literature" on the subject
                        thread.stop()
                        return

    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    finally:
        print("Stopping networking thread")
        thread.stop()
        print("Waiting for networking thread to stop")
        thread.join()



import json


def get_recent_block_heights(block_height):
    oldness = list(range(10)) + [pow(x, 2) for x in range(4, 64)]
    heights = [x for x in [block_height - o for o in oldness] if x >= 0]
    return heights


def load_peers():
    from .peer import DisconnectedRemotePeer
    try:
        db = [tuple(li) for li in json.loads(open("peers.json").read())]
    except Exception:
        db = []

    return {
        (host, port, direction): DisconnectedRemotePeer(host, port, direction, None) for (host, port, direction) in db
    }
